#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

import copy
from typing import Dict, Any, List, Union, Tuple
import datetime
import time
import networkx as nx
from pathlib import Path

import scine_database as db
from scine_database.energy_query_functions import get_energy_change, get_barriers_for_elementary_step_by_type
from scine_heron.multithread import Worker

from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.wrapped_label import WrappedLabel
from scine_heron.toolbar.io_toolbar import ToolBarWithSaveLoad
from scine_heron.statusbar.status_bar import StatusBar
from scine_heron.io.file_browser_popup import get_save_file_name, get_load_file_name
from scine_heron.settings.class_options_widget import ClassOptionsWidget, CompoundCostOptionsWidget
from scine_heron.containers.without_wheel_event import NoWheelDoubleSpinBox
from scine_heron.settings.dict_option_widget import DictOptionWidget
from scine_heron.settings.docstring_parser import DocStringParser

from scine_heron.database.graphics_items import Compound, Pathinfo, Reaction
from scine_heron.database.reaction_compound_view import ReactionAndCompoundView, ReactionAndCompoundViewSettings
from scine_heron.database.path_energy_levels import PathLevelWidget


from scine_chemoton.gears.pathfinder import Pathfinder as pf

from PySide2.QtWidgets import (
    QGraphicsItem,
    QGraphicsPathItem,
    QWidget,
    QCheckBox,
    QGraphicsView,
    QVBoxLayout,
    QHBoxLayout,
    QGraphicsScene,
    QLineEdit,
    QLabel,
    QScrollArea,
)
from PySide2.QtGui import QPainterPath, QGuiApplication, QPainter
from PySide2.QtCore import Qt, QPoint, QTimer, QThreadPool, SignalInstance


class CustomQScrollArea(QScrollArea):
    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if self.widget().mol_widget.underMouse() and modifiers == Qt.ControlModifier:
            self.widget().mol_widget.wheelEvent(event)
        elif self.widget().mol_widget.underMouse():
            super(CustomQScrollArea, self).wheelEvent(event)
        else:
            super(CustomQScrollArea, self).wheelEvent(event)


class GraphTravelWidget(QWidget):
    def __init__(self, parent: QWidget, db_manager: db.Manager, model: db.Model) -> None:
        self.window_width = 1440
        self.window_height = 810
        self.scene_object = QGraphicsScene(0, 0, self.window_width, self.window_height)
        super(GraphTravelWidget, self).__init__(parent=parent)
        self.db_manager = db_manager

        outer_layout_main = QVBoxLayout()
        inner_layout_main = QVBoxLayout()
        layout = QHBoxLayout()
        self.graph_travel_view = GraphTravelView(self, self.db_manager)
        self.settings = TraversalSettings(self, self.graph_travel_view, model)
        self.graph_travel_view.set_settings_widget(self.settings)

        self.settings_scroll_area = CustomQScrollArea()
        self.settings_scroll_area.setWidget(self.settings)
        self.settings_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.settings_scroll_area.setFixedWidth(self.settings._widget_width + 10)

        layout.addWidget(self.graph_travel_view)
        layout.addWidget(self.settings_scroll_area)

        inner_layout_main.addLayout(layout)
        inner_layout_main.setContentsMargins(9, 9, 9, 9)  # Default padding

        self.status_bar = StatusBar(self)
        self.settings.set_status_bar(self.status_bar)
        self.status_bar.update_status("Status:" + 4 * " " + "Ready", timer=None)

        outer_layout_main.addLayout(inner_layout_main)
        # Status bar
        layout_bottom = QHBoxLayout()
        layout_bottom.addWidget(self.status_bar)

        outer_layout_main.addLayout(layout_bottom)
        outer_layout_main.setContentsMargins(0, 0, 0, 0)  # Remove padding

        self.setLayout(outer_layout_main)

        self.view = QGraphicsView(self.scene_object)
        self.view.setLayout(outer_layout_main)
        self.view.setWindowTitle("Graph Traversal")
        # rendering smoother lines and edges of nodes
        self.view.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)

        self.view.show()


class GraphTravelView(ReactionAndCompoundView):
    def __init__(self, parent: QWidget, db_manager: db.Manager) -> None:
        super(GraphTravelView, self).__init__(parent=parent)
        self.setMouseTracking(True)
        self.setInteractive(True)

        # Add all data
        self.db_manager: db.Manager = db_manager
        self.line_items: Dict[str, Any] = {}

        self._curve_width, self._curve_height, self._curve_angle, self._curve_arc_length = 20, 20, 30, 60
        self._path_window = 140
        self._currently_plotting = False

        self.compound_start: str = ""
        self.compound_stop: str = ""

        self.pathinfo_list: List[Pathinfo] = []
        self._graph: nx.DiGraph = None

        # db-info of selected compound
        self._compound_collection = self.db_manager.get_collection("compounds")
        self._reaction_collection = self.db_manager.get_collection("reactions")
        self._flask_collection = self.db_manager.get_collection("flasks")
        self._elementary_step_collection = self.db_manager.get_collection("elementary_steps")
        self._property_collection = self.db_manager.get_collection("properties")
        self._structure_collection = self.db_manager.get_collection("structures")

    def remove_line_items(self):
        for line in self.line_items.values():
            self.scene_object.removeItem(line)
        # Reset line storage
        self.line_items = {}

    def remove_compound_items(self):
        for n_list in self.compounds.values():
            for n in n_list:
                self.scene_object.removeItem(n)
        # Reset node storage
        self.compounds = {}

    def remove_reaction_items(self):
        for n_list in self.reactions.values():
            for n in n_list:
                self.scene_object.removeItem(n)
        # Reset node storage
        self.reactions = {}

    def remove_pathinfo_items(self):
        for info in self.pathinfo_list:
            self.scene_object.removeItem(info)
        # Reset storage
        self.pathinfo_list = []

    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            zoomInFactor = 1.25
            zoomOutFactor = 1 / zoomInFactor

            # Save the scene pos
            oldPos = self.mapToScene(event.pos())

            # Zoom
            if event.angleDelta().y() > 0:
                zoomFactor = zoomInFactor
            else:
                zoomFactor = zoomOutFactor
            self.scale(zoomFactor, zoomFactor)
            # Center on mouse position
            self.centerOn(oldPos.x(), oldPos.y())
        elif event.modifiers() == Qt.ShiftModifier:
            self.horizontalScrollBar().wheelEvent(event)
        else:
            self.verticalScrollBar().wheelEvent(event)

    def plot_paths(self,
                   paths_excluding_rxn_double_cross: List[Tuple[List[str], float]],
                   paths_including_rxn_double_cross: List[Tuple[List[str], float]],
                   true_sight: bool):
        if self._currently_plotting:
            return
        self._currently_plotting = True
        # # # Remove old items from current scene
        self.remove_line_items()
        self.remove_compound_items()
        self.remove_reaction_items()
        self.remove_pathinfo_items()
        self.scene_object.update()

        # given input information
        assert paths_including_rxn_double_cross
        assert self.settings
        path_row = 0
        # # # Paths excluding double rxn crossing are easier to interpret for the operator
        if true_sight or len(paths_excluding_rxn_double_cross) == 0:
            for path in paths_including_rxn_double_cross:
                path_rank = self.settings._first_index_double_rxns
                self.construct_single_path_objects(path[0], path_rank, path_row, path[1])
                path_row += 1
        else:
            for path in paths_excluding_rxn_double_cross:
                path_rank = self.settings._first_index_no_double_rxns
                self.construct_single_path_objects(path[0], path_rank, path_row, path[1])
                path_row += 1

        # move nodes above lines
        self.move_to_foreground(self.compounds)
        self.move_to_foreground(self.reactions)

        # Bounding rectangle around current scene
        rect = self.scene().itemsBoundingRect()
        self.scene().setSceneRect(rect)
        # Fit scene rectangle in view and center on rectangle center
        self.fitInView(self.sceneRect(), Qt.KeepAspectRatio)
        self.centerOn(self.sceneRect().center().x(), self.sceneRect().center().y() - 20)
        self._currently_plotting = False

    def construct_single_path_objects(self, path, path_rank, path_row_index, path_length):
        widget_y_position = path_row_index * self._path_window
        # Require pathfinder graph, only connectivity
        path_graph = self.__construct_path_graph(path)
        path_graph_positions = self.__determine_node_positions(path, path_graph, widget_y_position)
        # NOTE: Access x pos value better
        # Build path info (rank and length)
        path_rank += path_row_index + 1
        path_length = round(path_length, 2)
        self.__build_pathinfo_item(
            - 6 * 70,
            widget_y_position,
            path, path_rank, path_length,
            text="Number\nLength")
        text_info = '#' + str(path_rank) + "\n" + str(path_length)
        self.__build_pathinfo_item(-3 * 70, widget_y_position, path, path_rank, path_length, text=text_info)

        # loop over positions and draw rxn nodes, edges and aggregates
        for node, position in path_graph_positions.items():
            # Draw Rxn Nodes and edges
            if ";0;" in node or ";1;" in node:
                rxn = db.Reaction(db.ID(node.split(";")[0]), self._reaction_collection)
                side = int(node.split(";")[1])
                rxn_item = self.__build_reaction_item(position[0], position[1], rxn, side)
                # Draw incoming edges of rxn node
                for e_count, edge in enumerate(path_graph.in_edges(node)):
                    line_id = "_".join(list(edge)) + "_" + str(e_count) + "_" + str(path_row_index)
                    edge_item = self.__build_edge(QPoint(path_graph_positions[edge[0]][0],
                                                         path_graph_positions[edge[0]][1]),
                                                  rxn_item.incoming())
                    self.line_items[line_id] = edge_item
                # Draw outgoing edges of rxn node
                for e_count, edge in enumerate(path_graph.out_edges(node)):
                    line_id = "_".join(list(edge)) + "_" + str(e_count) + "_" + str(path_row_index)
                    edge_item = self.__build_edge(rxn_item.outgoing(),
                                                  QPoint(path_graph_positions[edge[1]][0],
                                                         path_graph_positions[edge[1]][1]))
                    self.line_items[line_id] = edge_item
            else:
                # Draw Aggregate Nodes
                node_stripped = node.split(";")[0]
                if self._graph.nodes(data=True)[node_stripped]['type'] == db.CompoundOrFlask.COMPOUND.name:
                    a_type = db.CompoundOrFlask.COMPOUND
                    aggregate = db.Compound(db.ID(node_stripped), self._compound_collection)
                else:
                    a_type = db.CompoundOrFlask.FLASK
                    aggregate = db.Flask(db.ID(node_stripped), self._flask_collection)

                self.__build_compound_item(position[0], position[1], aggregate, a_type)

    def __construct_path_graph(self, path: List[str]) -> nx.DiGraph:
        # Derive edges from path
        path_edges = [(path[i], path[i + 1]) for i in range(0, len(path) - 1)]
        path_graph = nx.DiGraph()
        for node in path:
            path_graph.add_node(node)
        for edge in path_edges:
            path_graph.add_edge(edge[0], edge[1], weight=1.0)

        r_counter = 0
        p_counter = 0
        for rxn_node_index, node in enumerate([filtered_node for filtered_node in path if ";" in filtered_node]):
            # Add incoming reactants to complete path graph
            for edge in self._graph.in_edges(node):
                if edge != path_edges[2 * rxn_node_index]:
                    reactant_node = edge[0] + ";r" + str(r_counter) + ";"
                    path_graph.add_node(reactant_node)
                    path_graph.add_edge(reactant_node, node, weight=1.0)
                    r_counter += 1
            # Add outgoing products to complete path graph
            for edge in self._graph.out_edges(node):
                if edge != path_edges[2 * rxn_node_index + 1]:
                    product_node = edge[1] + ";p" + str(p_counter) + ";"
                    path_graph.add_node(product_node)
                    path_graph.add_edge(node, product_node, weight=1.0)
                    p_counter += 1

        return path_graph

    @staticmethod
    def __construct_model_string(model: db.Model) -> str:
        model_string = model.method
        if model.basis_set and model.basis_set.lower() != "none":
            model_string += "/" + model.basis_set
        if model.solvation and model.solvation.lower() != "none" and model.solvent.lower() != "any":
            model_string += "(" + model.solvent + ")"
        return model_string

    @staticmethod
    def __determine_node_positions(path: List[str],
                                   path_graph: nx.DiGraph, path_y_pos: float) -> Dict[str, Tuple[float, float]]:
        positions = {}
        path_edges = [(path[i], path[i + 1]) for i in range(0, len(path) - 1)]
        # NOTE: maybe set this somewhere better
        x_shift = 35.0
        y_shift = 25.0

        path_node_x_pos = 0.0
        for node in path:
            positions[node] = (path_node_x_pos, path_y_pos)
            path_node_x_pos += x_shift * 2

        reactant_edges = []
        prodcut_edges = []
        for rxn_node_index, node in enumerate([filtered_node for filtered_node in path if ";" in filtered_node]):
            # Position of reactants
            # Count minus one due to path
            for in_count, edge in enumerate(path_graph.in_edges(node)):
                if edge != path_edges[2 * rxn_node_index]:
                    r_node = edge[0]
                    positions[r_node] = (positions[node][0] - x_shift,
                                         positions[node][1] - y_shift * (in_count))
                    reactant_edges.append((r_node, node))

            # Position of products
            for out_count, edge in enumerate(path_graph.out_edges(node)):
                if edge != path_edges[2 * rxn_node_index + 1]:
                    p_node = edge[1]
                    positions[p_node] = (positions[node][0] + x_shift,
                                         positions[node][1] + y_shift * (out_count))
                    prodcut_edges.append((node, p_node))

        return positions

    # # # START BUILD ITEM FUNCTIONS

    def __build_compound_item(self, x: int, y: int, db_compound: Union[db.Compound, db.Flask],
                              a_type: db.CompoundOrFlask) -> Tuple[Compound, str]:
        compound_item = Compound(x, y, pen=self.compound_pen, brush=self.get_aggregate_brush(a_type))
        compound_item.db_representation = db_compound
        cid_string = db_compound.id().string()
        self.__bind_functions_to_object(compound_item)
        self.add_to_compound_list(compound_item)
        self.scene_object.addItem(compound_item)
        return compound_item, cid_string

    def __build_reaction_item(self, x: int, y: int, reaction: db.Reaction, side: int):
        assert self.settings
        assert self.settings.pathfinder  # mypy ...
        # Include structure model
        if self.settings.pathfinder.options.use_structure_model:
            structure_model = self.settings.pathfinder.options.structure_model
            if not isinstance(structure_model, db.Model):
                raise RuntimeError("Structure model not set")
        else:
            structure_model = self.settings.pathfinder.options.model
        # Get elementary step from graph or database
        es_from_graph = None
        node_id = reaction.id().string() + ";" + str(side) + ";"
        if "elementary_step_id" in self._graph.nodes(data=True)[node_id]:
            es_id = db.ID(
                self._graph.nodes(data=True)[node_id]["elementary_step_id"])
            es_from_graph = db.ElementaryStep(es_id, self._elementary_step_collection)

        if es_from_graph is None:
            es, barriers = self._get_elementary_step_of_reaction_from_db(
                reaction,
                self.settings.pathfinder.options.model,
                structure_model)
        else:
            es = es_from_graph
            # Get barrier of ES
            barriers = get_barriers_for_elementary_step_by_type(
                es,
                "electronic_energy",
                self.settings.pathfinder.options.model,
                self._structure_collection,
                self._property_collection)
        assert es
        reaction_item = Reaction(
            x, y, pen=self.reaction_pen,
            brush=self.get_reaction_brush(es.get_type()),
            # invert direction if side is 1
            invert_direction=bool(side)
        )
        reaction_item.db_representation = reaction
        reaction_item.assigned_es_id = es.id()

        struct = db.Structure(es.get_reactants()[0][0], self._structure_collection)
        reaction_item.energy_difference = get_energy_change(
            es, "electronic_energy", struct.get_model(), self._structure_collection, self._property_collection)
        if barriers[0] is not None and barriers[1] is not None:
            reaction_item.set_barriers(barriers)
        if es.has_spline():
            reaction_item.spline = es.get_spline()

        self.__bind_functions_to_object(reaction_item)
        self.add_to_reaction_list(reaction_item)
        self.scene_object.addItem(reaction_item)

        return reaction_item

    def __build_pathinfo_item(self, x: int, y: int, path: List[Any], path_rank: int, path_length: float, text: str):
        pathinfo_item = Pathinfo(x, y, path, path_rank, path_length, text)
        pathinfo_item.bind_mouse_double_click_function(self.path_info_mouse_double_press)

        self.scene_object.addItem(pathinfo_item)
        self.pathinfo_list.append(pathinfo_item)

    def __build_edge(self, start: QPoint, end: QPoint):
        path = QPainterPath(start)
        path.lineTo(end)
        # ID must be unique for scene!
        edge_item = QGraphicsPathItem(path)
        edge_item.setPen(self.path_pen)
        # ID must be unique for scene!
        self.scene_object.addItem(edge_item)
        return edge_item

    # # # END BUILD ITEM FUNCTIONS
    def path_info_mouse_double_press(self, _, item: QGraphicsItem) -> None:
        assert self.settings
        # Include structure model
        if self.settings.pathfinder.options.use_structure_model:
            structure_model = self.settings.pathfinder.options.structure_model
            if not isinstance(structure_model, db.Model):
                raise RuntimeError("Structure model not set")
        else:
            structure_model = self.settings.pathfinder.options.model

        level_list = [0.0]
        is_barrierless_list = []
        type_list = []
        for rxn_id in [rxn_node for rxn_node in item.path if ';' in rxn_node]:

            # Get elementary step from graph or database
            es_from_graph = None
            if "elementary_step_id" in self._graph.nodes(data=True)[rxn_id]:
                es_id = db.ID(
                    self._graph.nodes(data=True)[rxn_id]["elementary_step_id"])
                es_from_graph = db.ElementaryStep(es_id, self._elementary_step_collection)

            if es_from_graph is None:
                rxn = db.Reaction(db.ID(rxn_id.split(';')[0]), self._reaction_collection)
                es, barriers = self._get_elementary_step_of_reaction_from_db(
                    rxn,
                    self.settings.pathfinder.options.model,
                    structure_model)
            else:
                es = es_from_graph
                barriers = get_barriers_for_elementary_step_by_type(
                    es_from_graph,
                    "electronic_energy",
                    self.settings.pathfinder.options.model,
                    self._structure_collection,
                    self._property_collection
                )

            # Check faulty barriers
            if any(barrier is None for barrier in barriers) or es is None:
                raise RuntimeError("Barrier is None")

            es_type = es.get_type()
            if es_type == db.ElementaryStepType.BARRIERLESS:
                is_barrierless_list.append(True)
            else:
                is_barrierless_list.append(False)

            side_of_approach = int(rxn_id.split(';')[1])
            if side_of_approach == 0:
                tmp_up = barriers[0]
                tmp_down = barriers[1]
            else:
                tmp_up = barriers[1]
                tmp_down = barriers[0]
            level_list.append(level_list[-1] + tmp_up)  # type: ignore
            level_list.append(level_list[-1] - tmp_down)  # type: ignore
        # # # Type list
        for node_id in item.path:
            if node_id is None or item.path_rank is None or item.path_length is None:
                continue
            # skip rxn nodes
            if ';' in node_id:
                continue
            if self._graph.nodes(data=True)[node_id]['type'] == db.CompoundOrFlask.COMPOUND.name:
                aggregate_type = db.CompoundOrFlask.COMPOUND
            elif self._graph.nodes(data=True)[node_id]['type'] == db.CompoundOrFlask.FLASK.name:
                aggregate_type = db.CompoundOrFlask.FLASK
            else:
                raise RuntimeError("Unknown aggregate type")
            type_list.append(aggregate_type)
        overall_rxn = self.settings.pathfinder.get_overall_reaction_equation(item.path)

        # Model string construction
        es_method_info = self.__construct_model_string(self.settings.pathfinder.options.model)
        # Structure model string construction
        if self.settings.pathfinder.options.use_structure_model \
           and self.settings.pathfinder.options.structure_model != self.settings.pathfinder.options.model:
            structure_model = self.settings.pathfinder.options.structure_model
            es_method_info += " energies on " +\
                              self.__construct_model_string(self.settings.pathfinder.options.structure_model)
            es_method_info += " structures"

        path_level_widget = PathLevelWidget(parent=self, db_manager=self.db_manager,
                                            levels=level_list, barrierless=is_barrierless_list,
                                            type_list=type_list,
                                            path_info=(item.path_rank, item.path_length, item.path),
                                            rxn_equation=overall_rxn,
                                            es_method_info=es_method_info)

        path_level_widget.show()

    def __bind_functions_to_object(self, object: Any) -> None:
        object.bind_mouse_press_function(self.mouse_press_function)
        object.setAcceptHoverEvents(True)
        object.bind_hover_enter_function(self.hover_enter_function)
        object.bind_hover_leave_function(self.hover_leave_function)
        object.bind_menu_function(self.menu_function)


class TraversalSettings(ReactionAndCompoundViewSettings):
    def __init__(self, parent: QWidget, graph_travel_view: GraphTravelView, model: db.Model) -> None:
        super(TraversalSettings, self).__init__(parent=parent, parsed_layout=QVBoxLayout())

        # Elements of widget relative to width of ReactionProfile Canvas
        self._widget_width = 480
        # Fixed width for widget
        self.setMinimumWidth(self._widget_width)
        self.setMaximumWidth(self._widget_width)

        self.graph_travel_view = graph_travel_view
        self.status_bar: Union[None, QWidget] = None

        self.p_layout.addWidget(ToolBarWithSaveLoad(self._save, self._load, self,
                                                    hover_text_save="Save Graph", hover_text_load="Load Graph"))

        self.start_box_label = QLabel(self)
        self.start_box_label.setText("Start ID")
        self.start_box_label.resize(int(self._widget_width * 0.95), 20)
        self.start_box_label.setFixedHeight(20)
        self.p_layout.addWidget(self.start_box_label)

        self.start_id_text = QLineEdit(self)
        self.start_id_text.resize(int(self._widget_width * 0.95), 40)
        self.start_id_text.setText("")
        self.p_layout.addWidget(self.start_id_text)

        self.target_box_label = QLabel(self)
        self.target_box_label.setText("Target ID")
        self.target_box_label.resize(int(self._widget_width * 0.95), 20)
        self.target_box_label.setFixedHeight(20)
        self.p_layout.addWidget(self.target_box_label)

        self.target_id_text = QLineEdit(self)
        self.target_id_text.resize(int(self._widget_width * 0.95), 40)
        self.target_id_text.setText("")
        self.p_layout.addWidget(self.target_id_text)
        # Search layout box
        hbox1layout = QHBoxLayout()
        self.button_search = TextPushButton(
            "Search",
            lambda: self.trigger_thread_function(
                self._search_function,
                self._print_progress))
        hbox1layout.addWidget(self.button_search)

        self.button_prev = TextPushButton("Previous", self.__prev_function)
        self.button_prev.setDisabled(True)
        hbox1layout.addWidget(self.button_prev)

        self.button_next = TextPushButton("Next", lambda: self.trigger_thread_function(
            self.__next_function, self._print_progress))
        self.button_next.setDisabled(True)
        hbox1layout.addWidget(self.button_next)

        self.p_layout.addLayout(hbox1layout)
        # List of currently found paths and bool to indicate if new paths have been found
        self.paths_excluding_reaction_double_crossing: List[Tuple[List[str], float]] = []
        self.paths_including_reaction_double_crossing: List[Tuple[List[str], float]] = []

        self.new_paths = False
        self._first_index_no_double_rxns = 0
        self._first_index_double_rxns = 0
        self._searched_unique_paths = False
        self._skipped_paths_counter = 0
        self._pathfinder_options_changed = False
        self._pathfinder_graph_from_file = False

        # Timer for plotting newly found paths
        self.check_new_paths_timer = QTimer(self)
        # Timer might be causing segfaults if too short
        self.check_new_paths_timer.setInterval(1)
        self.check_new_paths_timer.timeout.connect(self.__plot_new_paths)  # pylint: disable=no-member
        self.check_new_paths_timer.start()

        self.pathfinder = pf(self.graph_travel_view.db_manager)
        # Model from CRNetworkView
        self.pathfinder.options.model = copy.deepcopy(model)
        self.pathfinder.options.structure_model = copy.deepcopy(model)

        self._start_conditions: Dict[str, float] = {}

        # buttons

        self.button_settings = TextPushButton("Pathfinder Settings", self._show_settings)
        self.p_layout.addWidget(self.button_settings)

        self.button_ccost = TextPushButton("Start conditions", self._show_start_conditions)
        self.p_layout.addWidget(self.button_ccost)

        self.button_calculate_compound_costs = TextPushButton(
            "Calculate Compound Costs", lambda: self.trigger_thread_function(
                self.__calculate_compound_costs_function, self._print_progress))
        self.p_layout.addWidget(self.button_calculate_compound_costs)

        self.buttons_to_deactivate = [self.button_search, self.button_next, self.button_prev,
                                      self.button_settings, self.button_ccost, self.button_calculate_compound_costs]
        self.buttons_to_reactivate: List[Any] = []

        self.save_to_svg_button = TextPushButton("Save SVG", self.graph_travel_view.save_svg)
        self.p_layout.addWidget(self.save_to_svg_button)

        # Search time layout
        search_time_layout = QHBoxLayout()
        search_time_layout.addWidget(WrappedLabel("Max. Search Time"))
        self.max_search_time = NoWheelDoubleSpinBox()
        self.max_search_time.setMinimum(1.0)
        self.max_search_time.setMaximum(3600.0)
        self.max_search_time.setValue(60.0)
        search_time_layout.addWidget(self.max_search_time)
        self.p_layout.addLayout(search_time_layout)

        self.unique_paths_cbox = QCheckBox("Show unique paths only")
        self.unique_paths_cbox.setChecked(True)
        self.p_layout.addWidget(self.unique_paths_cbox)

        self.true_sight_cbox = QCheckBox("Activate true sight")
        self.true_sight_cbox.setToolTip("Allow identical reaction\n to occur twice in shown paths.")
        self.p_layout.addWidget(self.true_sight_cbox)

        # Add molecule viewer and trajectory widgets to layout
        self.p_layout.addWidget(self.mol_widget)
        self.mol_widget.setMinimumWidth(self._widget_width)
        self.mol_widget.setMaximumWidth(self._widget_width)
        self.mol_widget.setMinimumHeight(350)

        self.p_layout.addWidget(self.es_mep_widget)
        self.es_mep_widget.setMinimumHeight(250)
        self.es_mep_widget.setMaximumHeight(300)
        self.es_mep_widget.setMinimumWidth(self._widget_width)
        self.es_mep_widget.setMaximumWidth(self._widget_width)

        self.setLayout(self.p_layout)
        self.show()

    def set_status_bar(self, status_widget: QWidget):
        self.status_bar = status_widget

    def _save(self):
        timestamp = datetime.datetime.now().strftime("%m%d-%H%M%S")
        handler = self.pathfinder.options.graph_handler

        filename = get_save_file_name(self, "graph." + handler + "." + timestamp, ['json'])
        if filename is None:
            return
        self.pathfinder.export_graph(filename)

    def _load(self):
        graph_filename = get_load_file_name(self, "graph", ['json'])
        if graph_filename is None:
            return
        compound_cost_filename = get_load_file_name(self, "compound_costs", ["json"])
        if compound_cost_filename is None:
            self.pathfinder.load_graph(graph_filename)
            self._pathfinder_graph_from_file = True
        else:
            self.pathfinder.load_graph(graph_filename, compound_cost_filename)
            self._pathfinder_graph_from_file = True
        # # # Link graph of travel view to graph of graph handler
        self.graph_travel_view._graph = self.pathfinder.graph_handler.graph

    def _print_progress(self, signal):
        if signal[0]:
            self.status_bar.clear_message()
            self.status_bar.update_status("Status:" + 4 * " " + signal[1], timer=None)
        # When signal is false, error status
        else:
            self.status_bar.clear_message()
            self.status_bar.update_error_status("Status:" + 4 * " " + signal[1], timer=5 * 1000)

    @staticmethod
    def trigger_thread_function(func, info_func):
        worker = Worker(func)
        worker.signals.running.connect(info_func)

        pool = QThreadPool.globalInstance()
        pool.start(worker)

    def __path_has_reaction_double_crossing(self, pathfinder_path_nodes: List[str]):
        # Extract ID of reaction nodes of pathfinder path
        tmp_rxn_list = [node.split(";")[0] for node in pathfinder_path_nodes if ";" in node]
        # True if every reaction only appears once, else False
        return len(tmp_rxn_list) != len(list(set(tmp_rxn_list)))

    def __clear_found_paths(self):
        self.paths_excluding_reaction_double_crossing = []
        self.paths_including_reaction_double_crossing = []

    def _build_graph_function(self):
        self.pathfinder.build_graph()

    def _show_settings(self):
        docstring_parser = DocStringParser()
        docstrings = docstring_parser.get_docstring_for_object_attrs(self.pathfinder.__class__.__name__,
                                                                     self.pathfinder.options)
        prev_options = copy.deepcopy(self.pathfinder.options)
        options = DictOptionWidget.get_attributes_of_object(self.pathfinder.options)
        setting_widget = ClassOptionsWidget(
            options, docstrings, parent=self, add_close_button=True,
            allow_removal=False,
        )
        # # # Other Defaults for barrierless weight widget
        barrierless_weight_widget = setting_widget._option_widget._option_widgets['barrierless_weight']
        barrierless_weight_widget.setMaximum(float(1e14))
        barrierless_weight_widget.setValue(1e12)
        # # #
        setting_widget.exec_()
        DictOptionWidget.set_attributes_to_object(self.pathfinder.options, options)
        self._pathfinder_options_changed = bool(prev_options != self.pathfinder.options)

    def _show_start_conditions(self):
        # can load stuff from json
        new_compound_costs = copy.deepcopy(self._start_conditions)
        ccost_widget = CompoundCostOptionsWidget(new_compound_costs, parent=self)
        # exec for looking
        ccost_widget.show()

    def _disable_buttons(self):
        # Disable buttons to avoid overload
        buttons_to_reactivate = []
        for button in self.buttons_to_deactivate:
            if button.isEnabled():
                buttons_to_reactivate.append(button)
                button.setDisabled(True)
        return buttons_to_reactivate

    def _show_path_level_widget(self):
        path_level_widget = PathLevelWidget(parent=self, db_manager=self.graph_travel_view.db_manager)
        path_level_widget.show()

    def _find_unique_paths_and_store_in_cache(self) -> bool:
        """
        Wrapper function for finding unique paths and storing the results in the cache
        for including double crossing of reactions and
        for excluding double crossing of reactions.

        Returns
        -------
        search_timed_out : bool
            Bool indicating if the search was interrupted by a time out.
        """
        # Counter for following loops
        path_counter = 0
        search_timed_out = False
        # Loop until 15 paths are found
        search_start_time = time.time()
        while (path_counter < 15):
            tmp_paths = self.pathfinder.find_unique_paths(
                self.start_id_text.text(), self.target_id_text.text(), 1)
            if len(tmp_paths) == 0:
                break
            # Regardless if valid or not, save the path here
            self.paths_including_reaction_double_crossing.append(tmp_paths[0])
            # Count only, if it is a path without rxn double crossing
            if not self.__path_has_reaction_double_crossing(tmp_paths[0][0]):
                self.paths_excluding_reaction_double_crossing.append(tmp_paths[0])
                path_counter += 1
            if not self.pathfinder._use_old_iterator:
                self.pathfinder._use_old_iterator = True
            if ((time.time() - search_start_time) > float(self.max_search_time.value())):
                search_timed_out = True
                break
        return search_timed_out

    def _find_paths_and_store_in_cache(self) -> bool:
        """
        Wrapper function for finding non-unique paths and storing the results in the cache
        for including double crossing of reactions and
        for excluding double crossing of reactions.

        Returns
        -------
        search_timed_out : bool
            Bool indicating if the search was interrupted by a time out.
        """
        # counter
        path_counter = 0
        search_timed_out = False
        analyzed_paths = 0
        search_start_time = time.time()
        while (path_counter < 15):
            tmp_paths = self.pathfinder.find_paths(self.start_id_text.text(), self.target_id_text.text(), 1,
                                                   n_skipped_paths=self._skipped_paths_counter)
            if len(tmp_paths) == 0:
                break
            self._skipped_paths_counter += 1
            analyzed_paths += 1
            # Regardless if valid or not, save the path here
            self.paths_including_reaction_double_crossing.append(tmp_paths[0])
            # Count only, if it is a valid path
            if not self.__path_has_reaction_double_crossing(tmp_paths[0][0]):
                self.paths_excluding_reaction_double_crossing.append(tmp_paths[0])
                path_counter += 1
            if ((time.time() - search_start_time) > float(self.max_search_time.value())):
                search_timed_out = True
                break
        return search_timed_out

    def _search_function(self, progress_callback: SignalInstance):
        progress_callback.emit((True, "Calculating routes"))
        self.buttons_to_reactivate = self._disable_buttons()

        if self.start_id_text.text() == "" or self.target_id_text.text() == "":
            progress_callback.emit((False, "Add a start ID and a target ID for your search."))
            return

        if self.pathfinder.graph_handler is None or (
                self._pathfinder_options_changed and not self._pathfinder_graph_from_file):
            progress_callback.emit((True, "Building graph"))
            self._build_graph_function()
            self.graph_travel_view._graph = self.pathfinder.graph_handler.graph
            progress_callback.emit((True, "Calculating routes"))
            # Reset bool for changed options
            if self._pathfinder_options_changed:
                self._pathfinder_options_changed = False
            assert self.pathfinder
        else:
            assert self.pathfinder
            assert self.pathfinder.graph_handler
        # Reset iterator in pathfinder
        self.pathfinder._use_old_iterator = False
        self._skipped_paths_counter = 0
        # Clear all paths found so far
        self.__clear_found_paths()

        search_timed_out = False

        if self.unique_paths_cbox.isChecked():
            # Loop until 15 paths are found
            search_timed_out = self._find_unique_paths_and_store_in_cache()
            # Activate next Button
            self.buttons_to_reactivate.append(self.button_next)
            # Set search type memory
            self._searched_unique_paths = True
        else:
            # Loop until 15 paths are found
            search_timed_out = self._find_paths_and_store_in_cache()
            self._searched_unique_paths = False
            self.buttons_to_reactivate.append(self.button_next)

        # Reset found paths and first path index
        self.new_paths = True
        self.pathfinder._use_old_iterator = True
        self._first_index_double_rxns = 0
        self._first_index_no_double_rxns = 0

        # Signal to Status Box
        status_message = "Ready"
        if search_timed_out:
            status_message += " -- with search time out"
        if len(self.paths_excluding_reaction_double_crossing) == 0:
            status_message += " -- True sight mode"
        progress_callback.emit((True, status_message))

    def _paths_in_cache(self) -> bool:
        paths_in_cache = False
        # Check paths including rxn double crossings if box is checked
        if self.true_sight_cbox.isChecked():
            if (len(self.paths_including_reaction_double_crossing) - self._first_index_double_rxns) > 15:
                paths_in_cache = True
        # Check paths excluding rxn double crossings if box is checked
        else:
            if (len(self.paths_excluding_reaction_double_crossing) - self._first_index_no_double_rxns) > 15:
                paths_in_cache = True
        return paths_in_cache

    def __next_function(self, progress_callback: SignalInstance):  # pylint: disable=unused-argument

        if self.start_id_text.text() == "" or self.target_id_text.text() == "":
            progress_callback.emit((False, "Add a start ID and a target ID for your search."))
            return

        progress_callback.emit((True, "Calculating routes"))
        self.buttons_to_reactivate = self._disable_buttons()
        assert self.pathfinder
        increase_indices = False
        search_timed_out = False
        # Check, if previous search was unique or not
        if self.unique_paths_cbox.isChecked() == self._searched_unique_paths:
            if self._paths_in_cache():
                increase_indices = True
            elif self.unique_paths_cbox.isChecked():
                search_timed_out = self._find_unique_paths_and_store_in_cache()
            else:
                # Loop until 15 paths are found
                search_timed_out = self._find_paths_and_store_in_cache()

            # # # Safety check for case no rxn excluding double crossing are found
            if len(self.paths_excluding_reaction_double_crossing) == 0 and\
               (len(self.paths_including_reaction_double_crossing) - self._first_index_double_rxns) > 15:
                increase_indices = True
            # # # Determine on how to increase indices
            if increase_indices:
                self._first_index_no_double_rxns += 15
                self._first_index_double_rxns += 15
            else:
                # # # Set indices
                self._first_index_no_double_rxns = 15 * \
                    int((len(self.paths_excluding_reaction_double_crossing) - 1) / 15)
                self._first_index_double_rxns = 15 * int((len(self.paths_including_reaction_double_crossing) - 1) / 15)
        # # # Redo search if any entry of checked box changed
        else:
            self.trigger_thread_function(self._search_function, self._print_progress)

        # Signal to Status Box
        status_message = "Ready"
        if search_timed_out:
            status_message += " -- with search time out"
        if len(self.paths_excluding_reaction_double_crossing) == 0:
            status_message += " -- True sight mode"
        progress_callback.emit((True, status_message))

        self.new_paths = True

    def __prev_function(self):
        if self.unique_paths_cbox.isChecked() == self._searched_unique_paths:
            self.buttons_to_reactivate = self._disable_buttons()
            # Reduce start index by 15
            self._first_index_no_double_rxns -= 15
            self._first_index_double_rxns -= 15
            self.new_paths = True
        else:
            self.trigger_thread_function(self._search_function, self._print_progress)

    def __plot_new_paths(self):
        if self.new_paths:
            # Get end index based on length of stored reactions
            if len(self.paths_excluding_reaction_double_crossing) - self._first_index_no_double_rxns >= 15:
                no_double_cross_step = 15
            else:
                no_double_cross_step = len(self.paths_excluding_reaction_double_crossing) % 15

            if len(self.paths_including_reaction_double_crossing) - self._first_index_double_rxns >= 15:
                double_cross_step = 15
            else:
                double_cross_step = len(self.paths_including_reaction_double_crossing) % 15

            self.graph_travel_view.plot_paths(
                # List of list of visited reaction nodes
                [path_data for path_data in
                    self.paths_excluding_reaction_double_crossing[
                        self._first_index_no_double_rxns:self._first_index_no_double_rxns + no_double_cross_step]],
                [path_data for path_data in
                    self.paths_including_reaction_double_crossing[
                        self._first_index_double_rxns:self._first_index_double_rxns + double_cross_step]],
                self.true_sight_cbox.isChecked()
            )

            # Add or remove previous button dependent on index
            if self._first_index_no_double_rxns > 0.0 or\
               (self.true_sight_cbox.isChecked() and self._first_index_double_rxns > 0.0):
                self.buttons_to_reactivate.append(self.button_prev)
            elif self.button_prev in self.buttons_to_reactivate:
                self.buttons_to_reactivate.remove(self.button_prev)
                self.button_prev.setDisabled(True)

            # Reactivate all of the required buttons
            for button in self.buttons_to_reactivate:
                button.setDisabled(False)
            self.buttons_to_reactivate = []
            # Reset trigger for plotting
            self.new_paths = False

    def __calculate_compound_costs_function(self, progress_callback: SignalInstance):  # pylint: disable=unused-argument
        if self.pathfinder is None:
            self.trigger_thread_function(self._build_graph_function, self._print_progress)
            assert self.pathfinder
        else:
            assert self.pathfinder
        progress_callback.emit((True, "Calculate Compound Costs"))
        # # # Set start conditions
        self.pathfinder.set_start_conditions(self._start_conditions)
        # # # Determine the compound costs
        self.pathfinder.calculate_compound_costs()
        # # # Save compound costs
        timestamp = datetime.datetime.now().strftime("%m%d-%H%M%S")
        filename = Path("compound_costs." + timestamp + ".json")
        self.pathfinder.export_compound_costs(filename)
        # # # Update the graph with the information of the compound costs
        self.pathfinder.update_graph_compound_costs()

        progress_callback.emit((True, "Ready"))
        if self.start_id_text.text() != "" and self.target_id_text.text() != "":
            self.trigger_thread_function(self._search_function,
                                         self._print_progress)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from copy import deepcopy
from pkgutil import iter_modules
from typing import Any, Dict, List, Optional

import numpy as np
import scine_sparrow  # noqa # pylint: disable=unused-import
import scine_utilities as su
from scine_utilities import settings_names

from scine_heron.mediator_potential.custom_results import CustomResult
from scine_heron.utilities import write_error_message, write_info_message, docstring_dict_from_scine_settings
from scine_heron.settings.docstring_parser import DocStringParser

# import all optional calculator modules if available
for _, name, __ in iter_modules():
    if name == "scine_ams_wrapper":
        try:
            import scine_ams_wrapper  # noqa # pylint: disable=(unused-import,import-error)
        except ImportError as exception:
            write_error_message(f"Could not import scine_ams_wrapper: {exception}")
    elif name == "scine_dftbplus_wrapper":
        try:
            import scine_dftbplus_wrapper  # noqa # pylint: disable=(unused-import,import-error)
        except ImportError as exception:
            write_error_message(f"Could not import scine_dftbplus_wrapper: {exception}")
    elif name == "scine_serenity":
        try:
            import scine_serenity_wrapper  # noqa # pylint: disable=(unused-import,import-error)
        except ImportError as exception:
            write_error_message(f"Could not import scine_serenity_wrapper: {exception}")
    elif name == "scine_swoose":
        try:
            import scine_swoose  # noqa # pylint: disable=(unused-import,import-error)
        except ImportError as exception:
            write_error_message(f"Could not import scine_swoose: {exception}")
    elif name == "scine_xtb_wrapper":
        try:
            import scine_xtb_wrapper  # noqa # pylint: disable=(unused-import,import-error)
        except ImportError as exception:
            write_error_message(f"Could not import scine_xtb_wrapper: {exception}")


class CalculatorLoadingFailed(Exception):
    pass


class ScineCalculatorWrapper:
    _calculator: Optional[su.core.Calculator]

    def __init__(self, method_family: str, program: str, hessian_required: bool,
                 settings: Optional[Dict[str, Any]] = None, atoms: Optional[su.AtomCollection] = None) -> None:
        self.__hessian_required = hessian_required
        self.__bond_orders_required = False
        self._docstring_parser = DocStringParser()
        success = self.load_calculator(method_family, program, atoms, settings)
        if not success:
            raise CalculatorLoadingFailed
        self.prev_molecule_version: Optional[int] = None if atoms is None else -1
        if success:
            self._set_required_properties()

    def get_structure(self) -> Optional[su.AtomCollection]:
        if self._calculator is not None and self.prev_molecule_version is not None:
            return self._calculator.structure
        return None

    def set_hessian_flag(self, value: bool) -> None:
        self.__hessian_required = value
        self._set_required_properties()

    def set_bond_orders_flag(self, value: bool) -> None:
        self.__bond_orders_required = value
        self._set_required_properties()

    def _set_required_properties(self) -> None:
        if self._calculator is None:
            return
        default_properties = [su.Property.Energy, su.Property.Gradients, su.Property.AtomicCharges,
                              su.Property.SuccessfulCalculation]
        additional_properties = []
        if self.__hessian_required:
            if self._calculator.get_possible_properties().contains_subset(su.PropertyList(su.Property.AtomicHessians)):
                additional_properties.append(su.Property.AtomicHessians)
            else:
                additional_properties.append(su.Property.Hessian)
        if self.__bond_orders_required:
            if not self._calculator.get_possible_properties().contains_subset(
                    su.PropertyList(su.Property.BondOrderMatrix)
            ):
                raise RuntimeError(f"Bond orders are not available for {self._calculator.name()}")
            additional_properties.append(su.Property.BondOrderMatrix)
        self._calculator.set_required_properties(default_properties + additional_properties)

    def calculate(self) -> su.Results:
        if self._calculator is None:
            return su.Results()
        return self._calculator.calculate()

    def calculate_custom_result(self) -> CustomResult:
        if self._calculator is None:
            return CustomResult()
        self._calculator.log = su.core.Log.silent()
        scine_results = su.Results()
        try:
            scine_results = self._calculator.calculate()
        except RuntimeError as exc:
            exception_to_raise: Optional[RuntimeError] = exc
            if self._multiplicity_convenience_check(exc):
                # could fix problem, by changing multiplicity, try again
                try:
                    scine_results = self._calculator.calculate()
                    exception_to_raise = None
                except RuntimeError as another_exc:
                    exception_to_raise = another_exc
            if exception_to_raise is not None:
                return self._failed_result(str(exception_to_raise))
        if not scine_results.successful_calculation:
            # if calculation did not converge, return empty
            return self._failed_result("Calculation was not successful")
        wf_generator = su.core.to_wf_generator(self._calculator)
        molden_input = "" if wf_generator is None else wf_generator.output_wavefunction()

        return CustomResult(
            result=scine_results,
            molden_input=molden_input,
            positions=self._calculator.structure.positions,
            error_msg="",
            info_msg="",
            settings=self._calculator.settings.as_dict()
        )

    def _failed_result(self, error: str) -> CustomResult:
        result = CustomResult()
        if self._calculator is None:
            return result
        result.error_msg = error
        result.settings = self._calculator.settings.as_dict()
        return result

    def update_system(self, molecule_version: int, elements: List[str], positions: np.ndarray,
                      settings: Optional[Dict[str, Any]] = None) -> None:
        if self._calculator is None:
            return
        if self.prev_molecule_version is None or molecule_version != self.prev_molecule_version:
            atoms = su.AtomCollection([su.ElementInfo.element_from_symbol(e) for e in elements], positions)
            self._calculator.structure = atoms
            self.prev_molecule_version = molecule_version
        else:
            self._calculator.positions = positions
        if settings is not None:
            self.update_settings(settings)

    def get_calculator(self) -> su.core.Calculator:
        if self._calculator is None:
            raise RuntimeError("Error retrieving calculator")
        return self._calculator

    def get_settings(self) -> su.Settings:
        if self._calculator is None:
            return su.Settings("empty", {})
        return self._calculator.settings

    def update_settings(self, new_settings: Dict[str, Any]) -> bool:
        if self._calculator is None:
            return False
        prev_settings = deepcopy(self._calculator.settings.as_dict())
        self._calculator.settings.update(new_settings)
        if not self._calculator.settings.valid():
            try:
                self._calculator.settings.throw_incorrect_settings()  # type: ignore
            except RuntimeError as e:
                write_error_message(str(e))
                self._calculator.settings.update(prev_settings)
                return False
        for key in new_settings.keys():
            if key not in self._calculator.settings.keys():
                write_error_message(f"Given setting {key} is not available for {self._calculator.name()}")
        return new_settings != prev_settings

    def load_calculator(self, method_family: str, program: str, atoms: Optional[su.AtomCollection] = None,
                        settings: Optional[Dict[str, Any]] = None) -> bool:
        try:
            calculator = su.core.get_calculator(method_family, program)
        except RuntimeError:
            write_error_message(f"Could not find calculator supporting {method_family} in {program} program")
            return False
        if calculator is None:
            write_error_message(f"Could not find calculator supporting {method_family} in {program} program")
            return False
        self._calculator = calculator
        if settings is not None:
            self.update_settings(settings)
        if atoms is not None:
            try:
                self._calculator.structure = atoms
            except RuntimeError as e:
                if self._multiplicity_convenience_check(e):
                    self._calculator.structure = atoms
                else:
                    write_error_message(str(e))
                    return False
        self._set_required_properties()
        self.prev_molecule_version = None
        return True

    def _multiplicity_convenience_check(self, exc: RuntimeError) -> bool:
        if self._calculator is not None and "chosen spin multiplicity" in str(exc):
            # assume we want to switch from even to uneven or vice versa
            current_mult: int = self._calculator.settings.get(settings_names.spin_multiplicity)  # type: ignore
            current_charge: int = self._calculator.settings.get(settings_names.molecular_charge)  # type: ignore
            new_min_mult = 1 if (current_mult % 2) == 0 else 2
            new_mult = max(current_mult - 1, new_min_mult)
            write_info_message(f"Loaded structure, set spin multiplicity to {new_mult} "
                               f"and molecular charge to {current_charge}")
            self._calculator.settings[settings_names.spin_multiplicity] = new_mult
            return True
        return False

    def get_docstring_dict(self) -> Dict[str, str]:
        if self._calculator is None:
            return {}
        return docstring_dict_from_scine_settings(self._calculator.settings)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the CentralWidget class.
"""

from PySide2.QtWidgets import QWidget, QVBoxLayout, QTabWidget
from PySide2.QtCore import QObject
from scine_heron.molecular_viewer import MolecularViewerWidget
from scine_heron.utilities import module_available
from typing import Optional


class CentralWidget(QWidget):
    """
    Sets up the central set of tabs.
    """

    def __init__(self, parent: QObject, haptic_client, method_family: Optional[str] = None,
                 program: Optional[str] = None):
        QWidget.__init__(self, parent)
        self.__layout = QVBoxLayout()
        self.tabs = QTabWidget()
        self.mol_view_tab = MolecularViewerWidget(self.tabs, haptic_client=haptic_client, method_family=method_family,
                                                  program=program)
        self.readuct_tab = QWidget()  # populated if ReaDuct is installed
        self.autocas_tab = QWidget()  # empty for now
        self.db_network_tab = QWidget()  # Empty, populated when DB is connected
        self.db_stats_tab = QWidget()  # Empty, populated when DB is connected
        self.db_viewer_tab = QWidget()  # Empty, populated when DB is connected
        self.chemoton_tab = QWidget()  # Empty, populated when DB is connected
        self.steering_tab = QWidget()  # Empty, populated when DB is connected
        self.kinetic_exploration_tab = QWidget()  # Empty, populated when DB is connected
        # Add tabs
        self.tabs.addTab(self.mol_view_tab, "Interactive")
        self.tabs.addTab(self.readuct_tab, "ReaDuct")
        self.tabs.addTab(self.autocas_tab, "AutoCAS")
        self.tabs.addTab(self.db_network_tab, "Reaction Network")
        self.tabs.addTab(self.db_stats_tab, "Network Statistics")
        self.tabs.addTab(self.db_viewer_tab, "Database Viewer")
        self.tabs.addTab(self.chemoton_tab, "Chemoton")
        self.tabs.addTab(self.steering_tab, "Steering Chemoton")
        self.tabs.addTab(self.kinetic_exploration_tab, "Kinetic Driven Exploration")
        # Disable tabs
        if module_available("scine_readuct"):
            from scine_heron.readuct.readuct_tab import ReaductTab
            self.readuct_tab = ReaductTab(self.tabs)
            self.tabs.removeTab(1)
            self.tabs.insertTab(1, self.readuct_tab, "Readuct")
        else:
            self.tabs.setTabEnabled(1, False)
            self.tabs.setTabVisible(1, False)
        if module_available("scine_autocas"):
            from scine_heron.autocas.autocas_widget import AutocasWidget
            self.autocas_tab = AutocasWidget(self.tabs)
            self.tabs.removeTab(2)
            self.tabs.insertTab(2, self.autocas_tab, "AutoCAS")
        else:
            self.tabs.setTabEnabled(2, False)
            self.tabs.setTabVisible(2, False)
        for i in range(3, 9):
            self.tabs.setTabVisible(i, False)
            self.tabs.setTabEnabled(i, False)
        # Add tabs to widget
        self.__layout.addWidget(self.tabs)
        self.setLayout(self.__layout)

    def toggle_db_dependent_tabs(
        self,
        compound_reaction: Optional[QWidget],
        database_monitor: Optional[QWidget],
        database_viewer: Optional[QWidget],
        chemoton_engines: Optional[QWidget],
        steering_wheel: Optional[QWidget],
        kinetic_exploration: Optional[QWidget],
    ) -> None:
        """
        Toggles the automatic update status and emits
        a signal if the status changed.
        """
        shift = int(module_available('scine_readuct')) + int(module_available("scine_autocas"))
        current = self.tabs.currentIndex()

        for index, (tab, member_attr, name) in enumerate(zip(
                [compound_reaction, database_monitor, database_viewer, chemoton_engines, steering_wheel,
                 kinetic_exploration],
                ["db_network_tab", "db_stats_tab", "db_viewer_tab", "chemoton_tab", "steering_tab",
                 "kinetic_exploration_tab"],
                ["Reaction Network", "Network Statistics", "Database Viewer", "Chemoton", "Steering Chemoton",
                 "Kinetics-Driven Exploration"]), 1 + shift):  # 1 because of Interactive
            if tab is not None:
                tab.setParent(self.tabs)
                setattr(self, member_attr, tab)
                self.tabs.removeTab(index)
                self.tabs.insertTab(index, tab, name)
                self.tabs.setTabEnabled(index, True)
                self.tabs.setTabVisible(index, True)
            else:
                self.tabs.setTabEnabled(index, False)
                self.tabs.setTabVisible(index, False)

        self.tabs.setCurrentIndex(current)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from datetime import datetime, timedelta
from os import path
from pkgutil import iter_modules
from typing import Any, Dict, Tuple, Optional, List

from PySide2.QtGui import (
    QGuiApplication,
    QClipboard,
    QBrush,
    QPen,
    QColor,
)
from PySide2.QtMultimedia import QMediaPlayer, QMediaContent
from PySide2.QtWidgets import (
    QScrollArea,
    QWidget,
    QSizePolicy
)
from PySide2.QtCore import Qt, QUrl

import scine_utilities as su

import scine_heron.config as config
from scine_heron.statusbar.status_bar import StatusBar


def copy_text_to_clipboard(text: str):
    clipboard = QGuiApplication.clipboard()
    clipboard.setText(text)
    if clipboard.supportsSelection():
        clipboard.setText(text, QClipboard.Selection)


def color_axis(axis):
    axis.xaxis.label.set_color(config.COLORS['primaryTextColor'])
    axis.yaxis.label.set_color(config.COLORS['primaryTextColor'])
    axis.tick_params(labelcolor=config.COLORS['primaryTextColor'])
    axis.set_facecolor(config.COLORS['secondaryLightColor'])
    axis.spines['bottom'].set_color(config.COLORS['primaryTextColor'])
    axis.spines['top'].set_color(config.COLORS['primaryTextColor'])
    axis.spines['right'].set_color(config.COLORS['primaryTextColor'])
    axis.spines['left'].set_color(config.COLORS['primaryTextColor'])


def get_font():
    return {"color": config.COLORS['primaryTextColor']}


def color_figure(figure):
    figure.patch.set_facecolor(config.COLORS['secondaryLightColor'])


def get_primary_line_color() -> str:
    return config.COLORS['primaryLineColor']


def get_secondary_line_color() -> str:
    return config.COLORS['secondaryLineColor']


def get_primary_light_color() -> str:
    return config.COLORS['primaryLightColor']


def get_secondary_light_color() -> str:
    return config.COLORS['secondaryLightColor']


def hex_to_qcolor(hex: str) -> QColor:
    return QColor(*(hex_to_rgb_base_255(hex)))


def hex_to_rgb_base_1(hex: str) -> Tuple[float, ...]:
    return tuple(x / 255.0 for x in hex_to_rgb_base_255(hex))


def hex_to_rgb_base_255(hex: str) -> Tuple[int, ...]:
    hex = hex.strip('#')
    assert len(hex) == 6
    return tuple(int(hex[i:i + 2], 16) for i in (0, 2, 4))


def qcolor_by_key(key: str) -> QColor:
    return hex_to_qcolor(config.COLORS[key])


def get_color_by_key(key: str) -> str:
    return config.COLORS[key]


def build_pen(
    color: QColor,
    style: Qt.PenStyle = Qt.PenStyle.SolidLine,
    join_style: Qt.PenJoinStyle = Qt.PenJoinStyle.RoundJoin,
    width: int = 1
) -> QPen:
    pen = QPen()
    pen.setStyle(style)
    pen.setJoinStyle(join_style)
    pen.setColor(color)
    pen.setWidth(width)
    return pen


def build_brush(fill: QColor, style: Qt.BrushStyle = Qt.BrushStyle.SolidPattern) -> QBrush:
    brush = QBrush()
    brush.setColor(fill)
    brush.setStyle(style)
    return brush


def datetime_to_query(time: datetime) -> Dict[str, Any]:
    return {'_lastmodified': {'$gt': {"$date": int(time.timestamp() * 1000)}}}


def _status_bar_impl() -> Optional[StatusBar]:
    from scine_heron import find_main_window
    from scine_heron.main_window import MainWindow
    main = find_main_window()
    if main is None or not isinstance(main, MainWindow):
        return None
    return main.get_status_bar()


def write_info_message(message: str, timer: Optional[int] = 10000) -> None:
    status_bar = _status_bar_impl()
    if status_bar is not None:
        status_bar.update_status(message, timer)


def write_error_message(message: str, timer: Optional[int] = 10000) -> None:
    status_bar = _status_bar_impl()
    if status_bar is not None:
        status_bar.update_error_status(message, timer)


def clear_status_bar() -> None:
    status_bar = _status_bar_impl()
    if status_bar is not None:
        status_bar.clear_message()


def vertical_scroll_area_wrap(content: QWidget) -> QScrollArea:
    scroll_area = QScrollArea()
    scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
    scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
    scroll_area.setWidget(content)
    scroll_area.setWidgetResizable(True)
    scroll_area.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
    return scroll_area


def construct_sound(sound: str) -> QMediaPlayer:
    from scine_heron import resource_path
    player = QMediaPlayer()
    player.setMedia(QMediaContent(QUrl.fromLocalFile(path.join(resource_path(), "sounds", f"{sound}.mp3"))))
    player.setVolume(100)
    return player


def docstring_dict_from_scine_settings(settings: su.Settings) -> Dict[str, str]:
    result = {}
    for key in settings.as_dict():
        result[key] = settings.descriptor_collection[key].property_description
    return result


def thread_safe_error(message: str, message_container: Optional[List[str]]):
    if message_container is None:
        write_error_message(message)
    else:
        message_container.append(message)


def module_available(module_name: str) -> bool:
    return module_name in (name for _, name, _ in iter_modules())


def timedelta_string(t: timedelta) -> str:
    """
    Custom string conversion for timedelta to get rid of microseconds from default __str__
    """
    new_t = t - timedelta(microseconds=t.microseconds)
    return str(new_t)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Iterator, List, Tuple, Union, Dict
from math import log
import networkx as nx
import numpy as np
from itertools import islice

import scine_database as db
from scine_database.concentration_query_functions import query_reaction_flux, query_concentration

from warnings import warn
warn('This is deprecated, please import Pathfinder from scine_chemoton', DeprecationWarning, stacklevel=2)


class Pathfinder:
    """
    A class to represent a list of reactions as graph and query this graph for simple paths between two nodes.
    In a simple path, every node part of the path is visited only once.

    Attributes
    ----------
    _calculations : db.Collection
        Collection of the calculations of the connected database.
    _compounds : db.Collection
        Collection of the compounds of the connected database.
    _reactions : db.Collection
        Collection of the reactions of the connected database.
    _elementary_steps : db.Collection
        Collection of the elementary steps of the connected database.
    _structures : db.Collection
        Collection of the structures of the connected database.
    _properties : db.Collection
        Collection of the properties of the connected database.
    graph_handler
        A class handling the construction of the graph. Can be adapted to ones needs.
    _use_old_iterator : bool
        Bool to indicate if the old iterator shall be used querying for paths between a source - target pair.
    _unique_iterator_memory : Tuple[str, str, Iterator]
        Memory of iterator with source and target string as well as the iterator.
    start_compounds : List[str]
        A list containing the compounds which are present at the start.
    start_compounds_set : bool
        Bool to indicate if start_compounds is set.
    _pseudo_inf : float
        Float for edges with infinite weight.
    compound_costs : Dict[str, float]
        A dictionary containing the cost of the compounds with the compounds as keys.
    compound_costs_solved : bool
        Bool to indicate if all compounds have a compound cost.
    """

    def __init__(self, db_manager: db.Manager):
        self.options = self.Options()
        self.manager = db_manager
        # Get required collections
        self._calculations = db_manager.get_collection('calculations')
        self._compounds = db_manager.get_collection('compounds')
        self._reactions = db_manager.get_collection('reactions')
        self._elementary_steps = db_manager.get_collection('elementary_steps')
        self._structures = db_manager.get_collection('structures')
        self._properties = db_manager.get_collection('properties')

        self.graph_handler: Union[Pathfinder.BasicHandler, None] = None
        # attribute to store iterator employed in find_unique_paths; path_object, iterator
        self._use_old_iterator = False
        self._unique_iterator_memory: Union[Tuple[Tuple[List[str], float],
                                                  Iterator], None] = None

        # Compound costs
        self.start_compounds: List[str] = []
        self.start_compounds_set = False
        self._pseudo_inf = 1e12
        self.compound_costs: Dict[str, float] = {}
        self.compound_costs_solved = False

    class Options:
        """
        A class to vary the setup of Pathfinder.
        """
        __slots__ = {"graph_handler", "barrierless_weight", "use_energy_threshold", "energy_threshold"}

        def __init__(self):
            self.graph_handler = "basic"  # pylint: disable=no-member
            self.barrierless_weight = 1.0
            # in kJ / mol
            self.use_energy_threshold = False
            self.energy_threshold = 100

    @staticmethod
    def get_valid_graph_handler_options() -> List[str]:
        return ["basic", "flux"]

    def _construct_graph_handler(self):
        """
        Constructor for the graph handler.

        Raises
        ------
        RuntimeError
            Invalid options for graph handler.
        """
        if not self.graph_handler:
            if self.options.graph_handler not in self.get_valid_graph_handler_options():
                raise RuntimeError("Invalid graph handler option.")
            if self.options.graph_handler == "basic":
                self.graph_handler = self.BasicHandler(self.manager)
                self.graph_handler.barrierless_weight = self.options.barrierless_weight
            elif self.options.graph_handler == "flux":
                self.graph_handler = self.FluxRxnAdder(self.manager)

    def _reset_iterator_memory(self):
        """
        Reset memory for unique memory.
        """
        self._unique_iterator_memory = None

    # Build graph function from reaction list
    def build_graph(self):
        """
        Build the nx.DiGraph() from a list of filtered reactions.
        """
        self._reset_iterator_memory()
        self._construct_graph_handler()
        assert self.graph_handler
        for rxn_id in self.graph_handler.get_valid_reaction_ids():
            rxn = db.Reaction(rxn_id, self._reactions)
            self.graph_handler.add_rxn(rxn)

    def find_paths(self, source: str, target: str, n_requested_paths: int = 3,
                   n_skipped_paths: int = 0) -> List[Tuple[List[str], float]]:
        """
        Query the build graph for simple paths between a source and target node.

        Notes
        -----
        * Requires a built graph

        Parameters
        ----------
        source : str
            The ID of the starting compound as string.
        target : str
            The ID of the targeted compound as string.
        n_requested_paths : int, optional
            Number of requested paths, by default 3
        n_skipped_paths : int, optional
            Number of skipped paths from, by default 0. Allows to set starting point of query.

        Returns
        -------
        found_paths : List[Tuple[List[str], float]]
            List of paths where each path consists of two information, the list of nodes of the path and its length.
        """
        assert self.graph_handler
        found_paths = []
        for path in self._k_shortest_paths(self.graph_handler.graph, source, target, n_requested_paths,
                                           weight="weight", path_start=n_skipped_paths):
            path_length = nx.path_weight(self.graph_handler.graph, path, "weight")
            found_paths.append((path, path_length))

        return found_paths

    def find_unique_paths(self, source: str, target: str, number: int = 3) -> List[Tuple[List[str], float]]:
        """
        Find a unique number of paths from a given source node to a given target node.
        Unique meaning that different paths (sequence of nodes) can have the same total weight.
        If one is solely interested in one path of paths with identical weight which can interpreted in the easiest
        human readable way, the shortest (in terms of length) longest (in terms of number of nodes) path is returned.
        Returning the path of a set of paths (maximum set size is set to 10) with the same weight with the most nodes
        guarantees that.

        Notes
        -----
        * Checks if a stored iterator for the given source-target pair should be used.

        Parameters
        ----------
        source : str
            The ID of the starting compound as string.
        target : str
            The ID of the targeted compound as string.
        number : int
            The number of unique paths to be returned. Per default, 3 paths are returned.

        Returns
        -------
        path_tuple_list : List[Tuple[List[str], float]]
            List of paths where each path consists of two information, the list of nodes of the path and its length.
        """
        assert self.graph_handler
        counter = 0
        path_tuple_list = list()
        # # # Initialise iterator over shortest simple paths if it is either not set or source/target do not match
        if not self._use_old_iterator or \
           self._unique_iterator_memory is None or \
           self._unique_iterator_memory[0][0][0] != source or \
           self._unique_iterator_memory[0][0][-1] != target:
            path_iterator = iter(nx.shortest_simple_paths(self.graph_handler.graph, source, target, weight="weight"))
            # # # Find first path and its cost
            old_path = next(path_iterator)
            old_path_cost = nx.path_weight(self.graph_handler.graph, old_path, weight="weight")
        else:
            path_iterator = self._unique_iterator_memory[1]
            old_path = self._unique_iterator_memory[0][0]
            old_path_cost = self._unique_iterator_memory[0][1]

        while counter < number:
            same_cost = True
            tmp_path_list: List[List[str]] = list()
            # # # Collect all paths with same cost
            n_max_collected_paths = 10
            while same_cost and len(tmp_path_list) < n_max_collected_paths:
                # # # Append old path to tmp_path list
                tmp_path_list.append(old_path)
                # # # Get next path and its cost
                new_path = next(path_iterator, None)
                # # # Break loop if no path is returned
                if new_path is None:
                    break
                new_path_cost = nx.path_weight(self.graph_handler.graph, new_path, weight="weight")
                # # # Check if new cost different to old cost
                if abs(old_path_cost - new_path_cost) > 1e-12:
                    same_cost = False
                # # # Overwrite old path with new path
                old_path = new_path

            # # # Append path with most nodes to tuple list and its cost
            path_tuple_list.append((max(tmp_path_list, key=lambda x: len(x)),  # pylint: disable=unnecessary-lambda
                                    old_path_cost))
            # # # Break counter loop if no more paths to target are found
            if new_path is None:
                break
            old_path_cost = new_path_cost
            counter += 1
        # # # Store iterator and path info (list of nodes and length)
        if new_path is not None:
            self._unique_iterator_memory = ((new_path, new_path_cost), path_iterator)
        return path_tuple_list

    @staticmethod
    def _k_shortest_paths(graph: nx.DiGraph, source: str, target: str, n_paths: int, weight: Union[str, None] = None,
                          path_start: int = 0) -> List[List[str]]:
        """
        Finding k shortest paths between a source and target node in a given graph.
        The length of the returned paths increases.

        Notes
        -----
        * This procedure is based on the algorithm by Jin Y. Yen [1]. Finding the first k paths requires O(k*nodes^3)
          operations.
        * Implemented as given in the documentation of NetworkX:
          https://networkx.org/documentation/stable/reference/algorithms/generated/ \
          networkx.algorithms.simple_paths.shortest_simple_paths.html
        * [1]: Jin Y. Yen, “Finding the K Shortest Loopless Paths in a Network”, Management Science, Vol. 17, No. 11,
               Theory Series (Jul., 1971), pp. 712-716.

        Parameters
        ----------
        graph : nx.DiGraph
            The graph to be queried.
        source : str
            The ID of the starting compound as string.
        target : str
            The ID of the targeted compound as string.
        n_paths : int
            The number of paths to be returned.
        weight : Union[str, None], optional
            The key for the weight encoded in the edges to be used.
        path_start : int, optional
            An index of the first returned path, by default 0

        Returns
        -------
        List[List[str]]
            List of paths, each path consisting of a list of nodes.
        """
        return list(
            islice(nx.shortest_simple_paths(graph, source, target, weight=weight), path_start, path_start + n_paths)
        )

    def set_start_conditions(self, conditions: Dict[str, float]):
        """
        Add the IDs of the start compounds to self.start_compounds and add entries for cost in self.compound_cost.

        Parameters
        ----------
        conditions : Dict
            The IDs of the compounds as keys and its given cost as values.
        """
        # # # Reset Start conditions, if already set previously
        if self.start_compounds_set:
            self.compound_costs = {}
            self.start_compounds = []
        # # # Loop over conditions and set them
        for cmp_id, cmp_cost in conditions.items():
            self.compound_costs[cmp_id] = cmp_cost
            self.start_compounds.append(cmp_id)
        self.start_compounds_set = True

    def __weight_func(self, u: str, _: str, d: Dict):
        """
        Calculates the weight of the edge d connecting u with _ (directional!).
        Only consider the costs of the required compounds if the edge d is from a compound node to a reaction node.
        If the edge d connects a reaction node with a compound node, the returned weight should be 0.

        Parameters
        ----------
        u : str
            The ID of start node.
        _ : str
            The ID of end node.
        d : Dict
            The edge connecting u and _ as dictionary

        Returns
        -------
        float
            Sum over the edge weight and the costs of the required compounds.
        """
        # # # Weight of edge
        edge_wt = d.get("weight", 0)
        # # # List of required compounds
        tmp_required_compounds = d.get("required_compounds", None)
        # # # Sum over costs of required compounds.
        # # # Only for edges from compound node to rxn node
        if ';' not in u and tmp_required_compounds is not None:
            required_compound_costs = np.sum([self.compound_costs[n] for n in tmp_required_compounds])
        else:
            required_compound_costs = 0.0

        return edge_wt + required_compound_costs

    def calculate_compound_costs(self, recursive: bool = True):
        """
        Determine the cost for all compounds via determining their shortest paths from the self.start_compounds.
        If this succeeds, set self.compound_costs_solved to True. Otherwise it stays False.

        The algorithm works as follows:
        Given the starting conditions, one loops over the individual starting compounds as long as:
            - the self._pseudo_inf entries in self.compound_costs are reduced
            - for any entry in self.compounds_cost a lower cost is found
        With each starting compound, one loops over compounds which have yet no cost assigned.
        For each start - target compound pair, the shortest path is determined employing Dijkstra's algorithm.
        The weight function checks the 'weight' of the edges as well as the costs of the required compounds listed in
        the 'required_compounds' of the traversed edges.
        If the path exceeds the length of self._pseudo_inf, this path is not considered for further evaluation.
        The weight of the starting compound is added to the tmp_cost.
        If the target compound has no weight assigned yet in 'self.compound_costs' OR
        if the target compound has a weight assigned which is larger (in 'self.compound_costs' as well as in
        'tmp_compound_costs') than the current 'tmp_cost' is written to the temporary storage of 'tmp_compound_costs'.

        After the loop over all starting compounds completes, the collected costs for the found targets are written to
        self.compound_costs.
        The convergence variables are updated and the while loop continues.

        Parameters
        ----------
        recursive : bool
            All compounds are checked for shorter paths, True by default.
            If set to False, compounds for which a cost has been determined are not checked in the next loop.
        """
        assert self.graph_handler
        if not self.start_compounds_set:
            raise RuntimeError("No start conditions given")
        if len(self.graph_handler.graph.nodes) == 0:
            raise RuntimeError("No nodes in graph")
        cmps_to_check = [n for n in self.graph_handler.graph.nodes if ';' not in n
                         and n not in self.start_compounds]
        # # # Set all compounds to be checked to pseudo inf cost
        for cmp_id in cmps_to_check:
            self.compound_costs[cmp_id] = self._pseudo_inf
        # # # Dictionary for current run
        tmp_compound_costs: Dict[str, float] = {}
        # # # Determine convergence variables for while loop
        current_inf_count = sum(value == self._pseudo_inf for value in self.compound_costs.values())
        prev_inf_count = current_inf_count
        # # # Convergence criteria
        compound_costs_change = None
        compound_costs_opt_change = 1
        converged = False
        # # # Find paths until either no costs are unknown or the None count has not changed
        while (not converged):
            compound_costs_opt_change = 0
            print("Remaining Nodes:", len(cmps_to_check))
            for tmp_start_node in self.start_compounds:
                # # # Loop over all nodes to be checked starting from start nodes
                for target in cmps_to_check:
                    # # # Determine cost and path with dijkstra's algorithm
                    tmp_cost, _ = nx.single_source_dijkstra(self.graph_handler.graph, tmp_start_node, target,
                                                            cutoff=None, weight=self.__weight_func)
                    # # # Check if the obtained cost is larger than infinity (pseudo_inf)
                    # # # and continue with the next target if this is the case
                    if (tmp_cost - self._pseudo_inf) > 0.0:
                        continue
                    # # # Add cost of the starting node
                    tmp_cost += self.compound_costs[tmp_start_node]
                    # # # Check for value in compound_costs dict and
                    if (self.compound_costs[target] != self._pseudo_inf and
                            10e-6 < self.compound_costs[target] - tmp_cost):
                        compound_costs_opt_change += 1
                    # # # Not already set check
                    if self.compound_costs[target] == self._pseudo_inf or (
                            self.compound_costs != self._pseudo_inf and 10e-6 < self.compound_costs[target] - tmp_cost):
                        # # # Not discovered in current run OR new tmp cost lower than stored cost
                        if (target not in tmp_compound_costs.keys()):
                            tmp_compound_costs[target] = tmp_cost
                        elif (tmp_cost < tmp_compound_costs[target]):
                            tmp_compound_costs[target] = tmp_cost
            # # # Write obtained compound_costs to overall dictionary
            for key, value in tmp_compound_costs.items():
                self.compound_costs[key] = value
                # # # Remove found nodes if recursive is false
                if not recursive:
                    cmps_to_check.remove(key)
            # # # Reset tmp_pr_cost
            tmp_compound_costs = {}
            # # # Convergence Check
            current_inf_count = sum(value == self._pseudo_inf for value in self.compound_costs.values())
            compound_costs_change = prev_inf_count - current_inf_count
            prev_inf_count = current_inf_count

            print(50 * '-')
            print("Current None:", current_inf_count)
            print("PR Change:", compound_costs_change)
            print("PR Opt Change:", compound_costs_opt_change)
            # # # Convergence Check
            if (compound_costs_change == 0 and compound_costs_opt_change == 0):
                converged = True

        # # # Final check if all compound costs are solved
        if current_inf_count == 0:
            self.compound_costs_solved = True

    def update_graph_compound_costs(self):
        """
        Update the 'weight' of edges from compound to reaction nodes by adding the compound costs.
        The compound costs are the sum over the costs stored in self.compound_costs of the required compounds.
        The edges of the resulting graph contain a weight including the required_compound_costs based on the starting
        conditions.
        All analysis of the graph therefore depend on the starting conditions.
        """

        # # # Check if all costs are available
        if not self.compound_costs_solved:
            unsolved_cmp = [key for key, _ in self.compound_costs.items()]
            raise RuntimeError("The following cmp have no cost assigned:\n" + str(unsolved_cmp) +
                               "\nReconsider the starting conditions.")
        # # # Reset unique_iterator_list as graph changes
        self._reset_iterator_memory()
        for node in self.compound_costs.keys():
            # # # Loop over all edges of compound and manipulate weight
            for target_node, attributes in self.graph_handler.graph[node].items():
                required_compound_costs = np.asarray([self.compound_costs[k] for k in attributes['required_compounds']])
                tot_required_compound_costs = np.sum(required_compound_costs)
                # # # Set required compound costs in edge
                self.graph_handler.graph.edges[node,
                                               target_node]['required_compound_costs'] = tot_required_compound_costs
                # # # Add required compound costs to weight
                self.graph_handler.graph.edges[node, target_node]['weight'] += tot_required_compound_costs

    # Base Class for adding a reaction; up weight to rxn node is just one per default, no further info
    class BasicHandler:
        """
        A basic class to handle the construction of the nx.DiGraph.
        A list of reactions can be added differently, depending on the implementation of _get_weight and
        get_valid_reaction_ids.
        """

        def __init__(self, manager: db.Manager):
            self.graph = nx.DiGraph()
            self.db_manager = manager
            self.barrierless_weight = 1.0
            self._compounds = self.db_manager.get_collection("compounds")
            self._flasks = self.db_manager.get_collection("flasks")
            self._structures = self.db_manager.get_collection("structures")
            self._properties = self.db_manager.get_collection("properties")
            self._reactions = self.db_manager.get_collection("reactions")

        def add_rxn(self, reaction: db.Reaction):
            """
            Add a reaction to the graph.
            Each reaction node represents the LHS and RHS.
            Hence every reagent of a reaction is connected to every product of a reaction via one reaction node.
            For instance:
            A + B = C + D, reaction R
            A -> R_1 -> C
            A -> R_1 -> D
            B -> R_1 -> C
            B -> R_1 -> D

            C -> R_2 -> A
            C -> R_2 -> B
            D -> R_2 -> A
            D -> R_2 -> B

            Representing this reaction in the graph yields 4 compound nodes,
            2 reaction nodes (same reaction) and 16 edges (2*2*2*2).
            The weights assigned to the edges depends on the _get_weight implementation.

            The edges from a compound node to a reaction node contain several information:
                weight: the weight of the edge
                        1 if the reaction is not barrierless, otherwise it is set to self.barrierless_weight
                required_compounds: the IDs of the other reagents of this side of the reaction in a list
                required_compound_costs: the sum over all compound costs of the compounds in the required_compounds list
                                         None by default.

            The edges from a reaction node to a compound node contain several information:
                weight: the weight of the edge, set to 0
                required_compounds: the IDs of the other products emerging;
                                        added for easier information extraction during the path analysis

            Parameters
            ----------
            reaction : db.Reaction
                The reaction to be added to the graph.
            """
            # Add two rxn nodes
            rxn_nodes = []
            reaction_id = reaction.id().string()

            for i in range(0, 2):
                # Add rxn node between lhs and rhs compound
                rxn_node = ';'.join([reaction_id, str(i)])
                rxn_node += ';'
                self.graph.add_node(rxn_node, color='rxn_node')
                rxn_nodes.append(rxn_node)
            # Convert to strings
            reactants = reaction.get_reactants(db.Side.BOTH)
            reactant_types = reaction.get_reactant_types(db.Side.BOTH)
            weights = self._get_weight(reaction)
            # Add lhs aggregates and connect
            for lhs_cmp, lhs_type in zip([i.string() for i in reactants[0]],
                                         [i.name for i in reactant_types[0]]):
                if lhs_cmp not in self.graph:
                    self.graph.add_node(lhs_cmp, type=lhs_type)
                required_cmps_lhs = [s.string() for s in reactants[0]]
                required_cmps_lhs.remove(lhs_cmp)
                self.graph.add_edge(lhs_cmp, rxn_nodes[0], weight=weights[0], required_compounds=required_cmps_lhs,
                                    required_compound_costs=None)
                self.graph.add_edge(rxn_nodes[1], lhs_cmp, weight=0.0, required_compounds=None)
            # Add rhs aggregates and connect
            for rhs_cmp, rhs_type in zip([i.string() for i in reactants[1]],
                                         [i.name for i in reactant_types[1]]):
                if rhs_cmp not in self.graph:
                    self.graph.add_node(rhs_cmp, type=rhs_type)
                required_cmps_rhs = [s.string() for s in reactants[1]]
                required_cmps_rhs.remove(rhs_cmp)
                self.graph.add_edge(rhs_cmp, rxn_nodes[1], weight=weights[1], required_compounds=required_cmps_rhs,
                                    required_compound_costs=None)
                self.graph.add_edge(rxn_nodes[0], rhs_cmp, weight=0.0, required_compounds=None)

            # # # Loop over reaction nodes to add required compounds info to downwards edges; might be unnecessary
            node_index = 1
            for node in rxn_nodes:
                for key in self.graph[node].keys():
                    self.graph.edges[node, key]['required_compounds'] = \
                        self.graph.edges[key, rxn_nodes[node_index]]['required_compounds']
                node_index -= 1

        def _get_weight(self, reaction: db.Reaction) -> Tuple[float, float]:
            """
            Determining the weights for the edges of the given reaction.

            Parameters
            ----------
            reaction : db.Reaction
                Reaction of interest

            Returns
            -------
            Tuple[float, float]
                Weight for connections to the LHS reaction node, weight for connections to the RHS reaction node
            """
            for step in reaction.get_elementary_steps(self.db_manager):
                # # # Barrierless weights for barrierless reactions
                if step.get_type() == db.ElementaryStepType.BARRIERLESS:
                    return self.barrierless_weight, self.barrierless_weight
            return 1.0, 1.0

        def get_valid_reaction_ids(self) -> List[db.ID]:
            """
            Basic filter function for reactions.
            Per default it returns all reactions.

            Returns
            -------
            List[db.ID]
                List of IDs of the filtered reactions.
            """
            valid_ids: List[db.ID] = list()
            for reaction in self._reactions.iterate_all_reactions():
                valid_ids.append(reaction.id())
            return valid_ids

    class FluxRxnAdder(BasicHandler):
        def __init__(self, manager: db.Manager):
            """
            Edge weights based on the the reaction flux from a kinetic modeling calculation.
            """
            super().__init__(manager)
            self._calculations = manager.get_collection("calculations")
            self._forward_flux_label = "_forward_edge_flux"
            self._backward_flux_label = "_backward_edge_flux"
            self._zero_break_scale = 1e-4
            self._kinetic_modeling_job_order = "kinetx_kinetic_modeling"
            self._reaction_id_key = "reaction_ids"

        def _valid_reaction(self, reaction: db.Reaction) -> bool:
            forward_flux = query_reaction_flux(self._forward_flux_label, reaction, self._compounds, self._flasks,
                                               self._structures, self._properties)
            return forward_flux > 0.0

        def _get_weight(self, reaction: db.Reaction) -> Tuple[float, float]:
            forward_flux = query_reaction_flux(self._forward_flux_label, reaction, self._compounds, self._flasks,
                                               self._structures, self._properties)
            backward_flux = query_reaction_flux(self._forward_flux_label, reaction, self._compounds, self._flasks,
                                                self._structures, self._properties)
            # catch zeros.
            forward_flux = max(forward_flux, 1e-6)
            backward_flux = max(backward_flux, 1e-6)
            # The flux is a direct measure of the conditional probability to traverse the reaction.
            # However, it is not normalized to be in [0, 1] but rather in [0, inf]. Therefore, we scale the
            # approximate probability by a factor that should (more or less) shift the flux into the interval [0, 1].
            p_forward = self._zero_break_scale * forward_flux
            p_backward = self._zero_break_scale * backward_flux
            # The information value (or surprise) of the event (traversal of the reaction) may then be given by
            # -log(p) (this is adapted from the definition of the Shannon entropy, i.e. S = <-log(p)>).
            return max(-log(p_forward), 0.0), max(-log(p_backward), 0.0)

        def get_valid_reaction_ids(self) -> List[db.ID]:
            last_calculation = self._get_last_completed_calculation()
            if last_calculation is not None:  # last_calculation may be None.
                reaction_str_ids: List[str] = last_calculation.get_setting(self._reaction_id_key)  # type: ignore
                if len(reaction_str_ids) > 0:
                    # The reactions must be valid. We quickly check the first one.
                    check_reaction = db.Reaction(db.ID(reaction_str_ids[0]), self._reactions)
                    if self._valid_reaction(check_reaction):
                        return [db.ID(str_id) for str_id in reaction_str_ids]
            return []

        def _get_last_completed_calculation(self) -> Union[None, db.Calculation]:
            start_structure = self._get_one_input_structure()
            if start_structure:
                calculation_ids = start_structure.get_calculations(self._kinetic_modeling_job_order)
                for calculation_id in reversed(calculation_ids):
                    calculation = db.Calculation(calculation_id, self._calculations)
                    if calculation.get_status() != db.Status.COMPLETE:
                        continue
                    return calculation
            return None

        def _get_one_input_structure(self) -> Union[db.Structure, None]:
            for compound in self._compounds.iterate_all_compounds():
                compound.link(self._compounds)
                centroid = db.Structure(compound.get_centroid(), self._structures)
                start_concentration = query_concentration("start_concentration", centroid, self._properties)
                if start_concentration > 0.0:
                    return centroid
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the HapticClient class.
"""

try:
    import scine_heron_haptic as suh
    from scine_heron.haptic.haptic_callback import HapticCallback
except ImportError:
    pass
except NameError:
    pass

import numpy as np
from vtk import (
    vtkAtom,
    vtkMolecule,
    vtkTransform,
    vtkMatrix4x4,
    vtkCamera,
)
from typing import List, Optional

from scine_utilities import (
    ANGSTROM_PER_BOHR,
    ElementInfo,
)


class HapticClient:
    def __init__(self) -> None:
        self.device_is_available: bool = False
        self.transform = vtkTransform()
        self.callback: Optional[HapticCallback] = None

    def init_haptic_device(self) -> None:
        try:
            self.haptic_device_manager = suh.HapticDeviceManager()  # pylint: disable=attribute-defined-outside-init
            self.device_is_available = self.haptic_device_manager.init_haptic_device()
        except NameError:
            print("No haptic device interface detected. Continuing without haptic feedback.")

        if self.device_is_available:
            self.callback = HapticCallback()  # type: ignore[assignment]
            self.haptic_device_manager.add_haptic_callback(self.callback)

    def exit_haptic_device(self) -> None:
        if self.device_is_available:
            self.haptic_device_manager.exit_haptic_device()

    def update_atom(
        self, atom_index: int, atom: vtkAtom, new_atom: bool = False
    ) -> None:
        if self.device_is_available:
            position = atom.GetPosition()
            radius = ElementInfo.covalent_radius(ElementInfo.element(atom.GetAtomicNumber())) * ANGSTROM_PER_BOHR

            atom = suh.AtomData(
                atom_index, position.GetX(), position.GetY(), position.GetZ(), radius
            )

            if new_atom:
                self.haptic_device_manager.add_atom(atom)
            else:
                self.haptic_device_manager.update_atom(atom)

    def update_molecule(self, molecule: vtkMolecule) -> None:
        if self.device_is_available:
            self.haptic_device_manager.clear_molecule()

            for atom_index in range(molecule.GetNumberOfAtoms()):
                self.update_atom(atom_index, molecule.GetAtom(atom_index), True)

    def update_transform_matrix(
        self, camera: vtkCamera, azimuth: float, elevation: float
    ) -> None:
        view_up = camera.GetViewUp()
        axis = [
            -camera.GetViewTransformMatrix().GetElement(0, 0),
            -camera.GetViewTransformMatrix().GetElement(0, 1),
            -camera.GetViewTransformMatrix().GetElement(0, 2),
        ]

        rotate_transform = vtkTransform()

        # azimuth
        rotate_transform.Identity()
        rotate_transform.RotateWXYZ(azimuth, view_up)

        # elevation
        rotate_transform.RotateWXYZ(elevation, axis)
        rotate_transform.Update()

        transform = vtkTransform()
        transform.Concatenate(rotate_transform)
        transform.Concatenate(self.transform)
        transform.Update()

        self.transform = transform
        self.__set_transformation_matrix()

    def __set_transformation_matrix(self) -> None:
        if self.device_is_available:
            vtk_matrix = self.transform.GetMatrix()

            invert_matrix = vtkMatrix4x4()
            vtkMatrix4x4.Invert(vtk_matrix, invert_matrix)

            self.haptic_device_manager.set_transformation_matrix(
                self.__matrix2list(vtk_matrix), self.__matrix2list(invert_matrix)
            )

    @staticmethod
    def __matrix2list(vtk_matrix: vtkMatrix4x4) -> List[float]:
        matrix = list()

        for i in range(0, 4):
            for j in range(0, 4):
                matrix.append(vtk_matrix.GetElement(i, j))

        return matrix

    def set_calc_gradient_in_loop(self, calc_gradient_in_loop: bool) -> None:
        if self.device_is_available:
            self.haptic_device_manager.set_calc_gradient_in_loop(calc_gradient_in_loop)

    def update_gradient(self, gradients: np.ndarray, force_scaling: float = 1.0,) -> None:
        if self.device_is_available:
            self.haptic_device_manager.update_gradient(
                [
                    force_scaling * elem for gradient in gradients for elem in gradient
                ]  # convert 2d gradient to 1d
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ServerProcess class.
"""

import socket
from typing import Optional
from multiprocessing import Process
from scine_heron.mediator_potential.mediator_server import run_server
from .clientserver import send_data


class ServerProcess:
    def __init__(self) -> None:
        self.is_server_running = False
        self.__server_process: Optional[Process] = None

    def start(self) -> None:
        if self.__server_process is None:
            self.__server_process = Process(target=run_server, args=(self.send_stop_signal,))
            self.__server_process.start()

    @staticmethod
    def send_stop_signal() -> None:
        stop_signal = True
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect(("127.0.0.1", 55145))
            send_data(data=[stop_signal, None, None, None, None, None, None], socket=s)

    def stop(self) -> None:
        if self.__server_process is not None and self.__server_process.is_alive():
            self.send_stop_signal()
        if self.__server_process is not None:
            self.__server_process.join()
            self.__server_process = None

    def terminate(self) -> None:
        if self.__server_process is not None:
            self.send_stop_signal()
            self.__server_process.join(timeout=2)
            if self.__server_process.is_alive():
                self.__server_process.terminate()
            self.__server_process.join()
            self.__server_process = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtWidgets import (
    QTabBar,
    QStylePainter,
    QStyleOptionTab,
    QStyle,
    QTabWidget,
)
from PySide2.QtCore import QPoint, QRect


class VerticalTabBar(QTabBar):
    """
    A tab bar that displays the tabs vertically instead of horizontally.
    """

    West = QTabWidget.West

    def tabSizeHint(self, index):
        s = QTabBar.tabSizeHint(self, index)
        s.transpose()
        return s

    def paintEvent(self, _event):
        painter = QStylePainter(self)
        opt = QStyleOptionTab()

        for i in range(self.count()):
            self.initStyleOption(opt, i)
            painter.drawControl(QStyle.CE_TabBarTabShape, opt)
            painter.save()

            s = opt.rect.size()
            s.transpose()
            r = QRect(QPoint(), s)
            r.moveCenter(opt.rect.center())
            opt.rect = r

            c = self.tabRect(i).center()
            painter.translate(c)
            painter.rotate(90)
            painter.translate(-c)
            painter.drawControl(QStyle.CE_TabBarTabLabel, opt)
            painter.restore()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the StatusBar class.
"""

from typing import Optional

from PySide2.QtWidgets import QStatusBar
from PySide2.QtCore import QObject, QTimer


class StatusBar(QStatusBar):
    """
    Provide errors and warnings.
    """

    def __init__(self, parent: QObject):
        QStatusBar.__init__(self, parent)
        self.__current_message = ""
        self.__current_message_is_error = False
        self.__default_style = self.styleSheet()
        self.__timer = QTimer()
        self.__timer.timeout.connect(self.reset_style)  # pylint: disable=no-member

    def update_error_status(self, line: str, timer: Optional[int] = 10000, clear_status: bool = False) -> None:
        if self.__current_message and self.__current_message_is_error and not clear_status:
            line = f"{self.__current_message}; {line}"
        self.__current_message_is_error = True
        self.clear_message()
        self.__timer.stop()
        self.setStyleSheet(
            "QStatusBar{padding-left:8px;background:rgba(179, 43, 43, 255);color:black;font-weight:bold;}"
        )
        self.__write_content(line, timer)
        if timer is not None:
            self.__timer.start(timer + 50)
        else:
            self.__timer.start(50)

    def update_status(self, line: str, timer: Optional[int] = 10000) -> None:
        if self.__current_message and not self.__current_message_is_error:
            line = f"{self.__current_message}; {line}"
        self.__current_message_is_error = False
        self.clear_message()
        self.__write_content(line, timer)

    def __write_content(self, line: str, timer: Optional[int] = 10000) -> None:
        if timer is not None:
            self.showMessage(line, timer)
        else:
            self.showMessage(line)
        self.__current_message = line
        self.repaint()

    def reset_style(self):
        self.__current_message = ""
        self.setStyleSheet(self.__default_style)

    def clear_message(self):
        super().clearMessage()
        self.reset_style()
        self.repaint()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ClassOptionsWidget class.
"""

from typing import Optional, Dict, Any, List, Callable, Type, TYPE_CHECKING
import inspect

from PySide2.QtGui import QCloseEvent, QKeySequence
from PySide2.QtCore import Qt, QObject
from PySide2.QtWidgets import (
    QVBoxLayout,
    QScrollArea,
    QDialog,
)
from scine_utilities import opt_settings_names, settings_names, ValueCollection

from scine_heron.containers.buttons import TextPushButton
from scine_heron.utilities import write_error_message
from scine_heron.settings.dict_option_widget import DictOptionWidget
from scine_heron.settings.docstring_parser import DocStringParser

if TYPE_CHECKING:
    from scine_database import Model
else:
    from scine_heron.dependencies.optional_import import importer
    Model = importer("scine_database", "Model")


class ClassOptionsWidget(QDialog):
    """
    A popup widget for editing the options of a class or a dictionary.
    This class is basically a wrapper around DictOptionWidget.
    """

    def __init__(
        self,
        options: Any,
        docstring: Optional[Dict[str, str]] = None,
        parent: Optional[QObject] = None,
        add_close_button: bool = True,
        allow_additions: bool = False,
        allow_removal: bool = True,
        addition_suggestions: Optional[List[str]] = None,
        suggestions_by_name: Optional[Dict[str, Dict[str, Callable]]] = None,
        value_type: Optional[Type] = None,
        keys_excluded_from_io: Optional[List[str]] = None,
    ) -> None:
        """
        Constructs the widget, but does not show the pop-up.
        The widget is constructed with the object to be altered by the user and
        some additional parameters concerning the layout and possibilities of the widget.

        Parameters
        ----------
        options : Any
            The object to be altered by the user.
        docstring : Optional[Dict[str, str]], optional
            A dictionary containing the docstrings for the individual attributes
            or key/value pairs. The docstrings are added as tooltips, by default None
        parent : Optional[QObject], optional
            The parent widget, by default None
        add_close_button : bool, optional
            If the pop-up should have a close button, by default True
        allow_additions : bool, optional
            If additional key/value pairs may be added, by default False
        allow_removal : bool, optional
            If key/value pairs are allowed to be removed, by default True
        addition_suggestions : Optional[List[str]], optional
            A list of strings that are suggested for autocompletion for new keys,
            when the user wants to add a key/value pair, by default None
        suggestions_by_name : Optional[Dict[str, Dict[str, Callable]]], optional
            A dictionary of suggestions for sub dictionaries in the created widget
            with the key representing the key to the sub dictionary and the values
            being a dictionary with the keys being the keys of the subdictionary
            and the value being functions to generate the suggestions, by default None
        keys_excluded_from_io : Optional[List[str]], optional
            A list of keys that should be excluded from saving and loading, by default None
        """
        super(ClassOptionsWidget, self).__init__(parent)

        self._options = options
        if isinstance(options, dict) or isinstance(options, ValueCollection):
            self._option_widget = DictOptionWidget(
                self._options,
                parent=self,
                docstring_dict=docstring,
                add_close_button=False,
                allow_additions=allow_additions,
                allow_removal=allow_removal,
                addition_suggestions=addition_suggestions,
                suggestions_by_name=suggestions_by_name,
                value_type=value_type,
                keys_excluded_from_io=keys_excluded_from_io
            )
        else:
            self._option_widget = DictOptionWidget(
                DictOptionWidget.get_attributes_of_object(self._options),
                parent=self,
                docstring_dict=docstring,
                add_close_button=False,
                allow_additions=allow_additions,
                allow_removal=allow_removal,
                addition_suggestions=addition_suggestions,
                suggestions_by_name=suggestions_by_name,
                keys_excluded_from_io=keys_excluded_from_io
            )

        self.setWindowTitle("Options")
        self.setMinimumWidth(350)
        self.setMinimumHeight(200)

        layout = QVBoxLayout()

        self._scroll_area = QScrollArea()
        self._scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self._scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self._scroll_area.setWidget(self._option_widget)
        self._scroll_area.setWidgetResizable(True)
        layout.addWidget(self._scroll_area)

        if add_close_button:
            close_button = TextPushButton("Ok", self.reject)
            close_button.setShortcut(QKeySequence("Return"))
            layout.addWidget(close_button)

        self.setLayout(layout)

    def closeEvent(self, event: QCloseEvent) -> None:
        self._update_options()
        super().closeEvent(event)

    def reject(self) -> None:
        self._update_options()
        super().reject()

    def _update_options(self):
        DictOptionWidget.set_attributes_to_object(
            self._options, self._option_widget.get_widget_data()
        )


class ModelOptionsWidget(ClassOptionsWidget):
    """
    A widget that holds a model and allows to manipulate it.
    """

    def __init__(self, parent: Optional[QObject] = None, model: Optional[Model] = None):
        """
        Construct the widget with an optional starting model.

        Parameters
        ----------
        parent : Optional[QObject], optional
            The parent widget, by default None
        model : Optional[Model], optional
            An optional starting model, a simple PM6 model, if none is given, by default None
        """
        model = Model("PM6", "PM6", "") if model is None else model
        super().__init__(options=model, parent=parent, add_close_button=False, allow_additions=False,
                         allow_removal=False)

    @property
    def model(self) -> Model:
        """
        Getter for the underlying model
        """
        self._update_options()
        return self._options

    @model.setter
    def model(self, model: Model) -> None:
        """
        Setter for the underlying model.

        Parameters
        ----------
        model : Model
            The new model
        """
        self._options = model
        self._option_widget.setParent(None)  # type: ignore
        self._option_widget = DictOptionWidget(
            DictOptionWidget.get_attributes_of_object(self._options),
            parent=self.parent(),
            add_close_button=False,
            allow_additions=False,
            allow_removal=False,
        )
        self._scroll_area.setWidget(self._option_widget)


class GeneralSettingsWidget(ClassOptionsWidget):
    """
    A widget that is meant to build Scine calculator and ReaDuct task settings.
    I can also have a button that allows to add default Chemoton settings.
    """

    def __init__(self, parent: Optional[QObject] = None, pre_settings: Optional[Dict[str, Any]] = None,
                 add_chemoton_settings_button: bool = True, add_opt_suggestions: bool = True):
        """
        Construct the widget with some existing settings and if the Chemoton button
        should be added.

        Parameters
        ----------
        parent : Optional[QObject], optional
            The parent widget, by default None
        pre_settings : Optional[Dict[str, Any]], optional
            Some settings to add to the widget already, by default None
        add_chemoton_settings_button : bool, optional
            If the Chemoton settings button should be added, by default True
        add_opt_suggestions : bool, optional
            If the Scine Utilities optimization settings names should be added as suggestions
        """
        settings = {} if pre_settings is None else pre_settings
        names = [s for s in dir(settings_names) if not s.startswith("_")]
        if add_opt_suggestions:
            names += [s for s in dir(opt_settings_names) if not s.startswith("_")]
        super().__init__(options=settings, parent=parent, add_close_button=False, allow_additions=True,
                         addition_suggestions=names)
        if add_chemoton_settings_button:
            self.default_button = TextPushButton("Add default Chemoton NT Settings",
                                                 self.__add_chemoton_defaults)
            self.layout().addWidget(self.default_button)

    def remove_chemoton_button(self):
        """
        remove the button that allows to add Chemoton settings
        """
        self.default_button.setParent(None)  # type: ignore

    @property
    def settings(self) -> ValueCollection:
        """
        Getter for the underlying settings

        Returns
        -------
        ValueCollection
            The settings
        """
        self._options = {}
        self._update_options()
        return ValueCollection(self._options)

    def __add_chemoton_defaults(self) -> None:
        """
        Adds the Chemoton defaults to the settings
        """
        from scine_chemoton.default_settings import default_nt_settings
        self._update_options()
        defaults = default_nt_settings().as_dict()
        for k, v in defaults.items():
            if k not in self._options:
                self._option_widget.add_key_value(k, v)
        self._options = {**self._options, **defaults}


def generate_instance_based_on_potential_widget_input(calling_widget: QObject, cls: type,
                                                      predefined_kwargs: Optional[Dict[str, Any]] = None,
                                                      possible_suggestions: Optional[Dict[str,
                                                                                          Dict[str, Callable]]] =
                                                      None) -> Any:
    """
    Generate an instance of a given class type based on some user input if required

    Parameters
    ----------
    calling_widget : QObject
        The widget that called this function
    cls : type
        The class type to be instantiated
    predefined_kwargs : Optional[Dict[str, Any]], optional
        A dictionary containing the parameter name and the value that should be used for constructing the
        instance instead of asking the user in the widget, by default None
    possible_suggestions : Optional[Dict[str, Dict[str, Callable]]], optional
        Provided suggestions to be propagated to the ClassOptionsWidgets;
        A dictionary of suggestions for sub dictionaries in the created widget
        with the key representing the key to the sub dictionary and the values
        being a dictionary with the keys being the keys of the subdictionary
        and the value being functions to generate the suggestions, by default None

    Notes
    -----
    Currently not thread safe due to written error messages

    Returns
    -------
    Any
        The instance of the class or None if an error occurred
    """
    delete_list = ["self", "args", "kwargs", "_"]
    signature = inspect.signature(cls.__init__)  # type: ignore
    if len(signature.parameters) > 1:
        parameters = signature.parameters.copy()
        for d in delete_list:
            if d in parameters:
                del parameters[d]
        if predefined_kwargs is not None:
            for key in predefined_kwargs:
                if key not in parameters:
                    write_error_message(f"Failed to construct class '{cls.__name__}', "
                                        f"due to wrong predefined argument {key}")
                    return None
                del parameters[key]
            if not parameters:
                # nothing left for the user to decide, all was predefined
                return cls(**predefined_kwargs)
        elif not parameters:
            # can happen if the class was written inheritance friendly with empty arguments
            return cls()
        try:
            parser = DocStringParser()
            doc = parser.get_docstring_for_class_init(cls.__name__, cls)
        except BaseException:
            # print(str(e))
            doc = None  # pylint: disable=bad-option-value
        param_selection = ClassOptionsWidget(parameters, docstring=doc, parent=calling_widget, allow_removal=False,
                                             add_close_button=True, suggestions_by_name=possible_suggestions)
        param_selection.exec_()
        if predefined_kwargs is not None:
            parameters.update(predefined_kwargs)
        if any(isinstance(p, inspect.Parameter) for p in parameters.values()):
            write_error_message(f"Failed to construct class {cls.__name__}")
            return None
        for k, v in parameters.items():
            if isinstance(v, str):
                if v.startswith("[") and v.endswith("]"):
                    str_list = [s.strip() for s in v.replace("[", "").replace("]", "").split(",")]
                    if str_list and not (len(str_list) == 1 and isinstance(str_list[0], str) and not str_list[0]):
                        parameters[k] = str_list
        return cls(**parameters)
    return cls()


# # # Compound Cost Widget probably not wanted here
class CompoundCostOptionsWidget(ClassOptionsWidget):

    def __init__(
        self,
        options: Any,
        parent: QObject,
    ) -> None:
        super(CompoundCostOptionsWidget, self).__init__(options, parent=parent, allow_additions=True,
                                                        allow_removal=True, value_type=float)
        # Allow access to parent
        self.parent: QObject = parent

        # Deactivate buttons from parent and remember state before disabling
        self._buttons_disabled_state: List[bool] = []
        if hasattr(parent, 'buttons_to_deactivate'):
            for button in parent.buttons_to_deactivate:
                self._buttons_disabled_state.append(not button.isEnabled())
                button.setDisabled(True)

        self.setWindowTitle("Start conditions")
        self.setMinimumWidth(400)
        self.setMinimumHeight(200)

    def closeEvent(self, event: QCloseEvent) -> None:
        self._update_options()
        # Enable buttons as before opening this widget
        self._enable_buttons()

        super().closeEvent(event)

    def reject(self) -> None:
        self._update_options()
        # Enable buttons as before opening this widget
        self._enable_buttons()

        # assert self.parent.pathfinder
        if self._options != self.parent.pathfinder.options:
            self.parent._start_conditions = self._options
        super().reject()

    def _enable_buttons(self):
        if hasattr(self.parent, 'buttons_to_deactivate'):
            for button, disable in zip(self.parent.buttons_to_deactivate, self._buttons_disabled_state):
                button.setDisabled(disable)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Any, List, Optional, Tuple, TYPE_CHECKING, Dict, Union
import pickle

from PySide2.QtCore import QObject
from PySide2.QtWidgets import QWidget, QDialog, QLineEdit

from scine_chemoton.filters.aggregate_filters import (
    AggregateFilter, PlaceHolderAggregateFilter
)
from scine_chemoton.filters.reactive_site_filters import (
    ReactiveSiteFilter,
)
from scine_chemoton.filters.further_exploration_filters import (
    FurtherExplorationFilter,
)
from scine_chemoton.filters.elementary_step_filters import (
    ElementaryStepFilter, PlaceHolderElementaryStepFilter
)
from scine_chemoton.filters.reaction_filters import (
    ReactionFilter, PlaceHolderReactionFilter
)
from scine_chemoton.gears.network_refinement.enabling import (
    AggregateEnabling, ReactionEnabling, EnableCalculationResults, PlaceHolderAggregateEnabling,
    PlaceHolderReactionEnabling, PlaceHolderCalculationEnabling
)
from scine_chemoton.gears.network_refinement.disabling import (
    ReactionDisabling, StepDisabling
)
from scine_chemoton.filters.structure_filters import StructureFilter
from scine_heron.chemoton.class_selection_widget import ClassSelectionWidget, SelectionState
from scine_heron.containers.combo_box_tab_widget import ComboBoxTabWidget
from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.layouts import VerticalLayout
from scine_heron.io.json_pickle_wrap import encode, decode
from scine_heron.io.file_browser_popup import get_load_file_name, get_save_file_name
from scine_heron.toolbar.io_toolbar import ToolBarWithSaveLoad
from scine_heron.utilities import write_error_message, write_info_message

if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class AggregateFilterBuilderDialog(QDialog):

    filter_signal = Signal(AggregateFilter)
    reset_filter_signal = Signal()
    state_signal = Signal(SelectionState)

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)
        self.setWindowTitle("Specify Aggregate Filter")
        self._builder = FilterBuilder(parent=self, aggregate_filter_only=True)
        self._ok_button = TextPushButton("OK", self.accept, parent=self)
        self.setLayout(VerticalLayout([self._builder, self._ok_button]))

    def set_state(self, state: SelectionState) -> None:
        self._builder.set_states([state])

    def reject(self) -> None:
        self._send_signals()
        super().reject()

    def accept(self) -> None:
        self._send_signals()
        super().accept()

    def _send_signals(self) -> None:
        aggregate_filter = self._builder.get_aggregate_filter()
        if aggregate_filter is None:
            self.reset_filter_signal.emit()
        else:
            self.filter_signal.emit(aggregate_filter)
            self.state_signal.emit(self._builder.get_states()[0])


class AggregateFilterBuilderButton(TextPushButton):

    def __init__(self, parent: Optional[QWidget] = None, shortcut: Optional[str] = None) -> None:
        super().__init__("Specify Filters", self._pop_up, parent=parent, shortcut=shortcut)
        self._aggregate_filter = AggregateFilter()
        self._last_state = SelectionState()

    def get_aggregate_filter(self) -> AggregateFilter:
        return self._aggregate_filter

    def reset_filter(self) -> None:
        write_error_message("AggregateFilter could not be built")
        self._aggregate_filter = FilterBuilder.default_filters()[0]
        self._last_state = SelectionState()

    def set_aggregate_filter(self, aggregate_filter: AggregateFilter) -> None:
        self._aggregate_filter = aggregate_filter

    def set_state(self, state: SelectionState) -> None:
        self._last_state = state

    def _pop_up(self) -> None:
        dialog = AggregateFilterBuilderDialog(parent=self)
        dialog.set_state(self._last_state)
        dialog.reset_filter_signal.connect(self.reset_filter)  # pylint: disable=no-member
        dialog.filter_signal.connect(self.set_aggregate_filter)  # pylint: disable=no-member
        dialog.state_signal.connect(self.set_state)  # pylint: disable=no-member
        dialog.exec_()


class FilterBuilder(ComboBoxTabWidget):
    """
    A container widget based on tabs with each tab being a ClassSelectionWidget
    for the individual filter types.
    """

    def __init__(self, parent: Optional[QObject] = None, aggregate_filter_only: bool = False) -> None:
        """
        Construct the filter builder with currently hardcoded tabs.

        Parameters
        ----------
        parent : Optional[QObject], optional
            The parent widget, by default None
        """
        # first call super with None, then set tabs, so we can set ourselves as parent
        super().__init__(tabs=None, parent=parent)
        # save load
        self._layout.addWidget(ToolBarWithSaveLoad(self._save, self._load, parent=self))
        # hard code the tabs to be given to our parent to set them up
        self._agg_widget = ClassSelectionWidget(AggregateFilter, "filter", parent=self)
        self._tabs: Dict[str, Union[ClassSelectionWidget, QLineEdit, None]] = {
            "Aggregate Filter": self._agg_widget,
        }
        if aggregate_filter_only:
            self._site_widget: Optional[ClassSelectionWidget] = None
            self._further_widget: Optional[ClassSelectionWidget] = None
            self._structure_widget: Optional[ClassSelectionWidget] = None
            self._elementary_step_widget: Optional[ClassSelectionWidget] = None
            self._reaction_widget: Optional[ClassSelectionWidget] = None
            self._aggregate_enabling_widget: Optional[ClassSelectionWidget] = None
            self._aggregate_validation_widget: Optional[ClassSelectionWidget] = None
            self._reaction_enabling_widget: Optional[ClassSelectionWidget] = None
            self._reaction_disabling_widget: Optional[ClassSelectionWidget] = None
            self._step_disabling_widget: Optional[ClassSelectionWidget] = None
            self._reaction_validation_widget: Optional[ClassSelectionWidget] = None
            self._results_enabling_widget: Optional[ClassSelectionWidget] = None
        else:
            self._site_widget = ClassSelectionWidget(ReactiveSiteFilter, "filter", parent=self)
            self._further_widget = ClassSelectionWidget(FurtherExplorationFilter, "filter", parent=self)
            self._structure_widget = ClassSelectionWidget(StructureFilter, "filter", parent=self)
            self._elementary_step_widget = ClassSelectionWidget(ElementaryStepFilter, "filter", parent=self)
            self._reaction_widget = ClassSelectionWidget(ReactionFilter, "filter", parent=self)
            # TODO: Hide the enabling/disabling logic in the GUI for the time being
            #  because this requires the implementation of additional widgets to construct
            #  the enabling/disabling objects.
            info = QLineEdit("Not supported yet")
            info.setReadOnly(True)
            self._aggregate_enabling_widget = info
            self._aggregate_validation_widget = info
            self._reaction_enabling_widget = info
            self._reaction_disabling_widget = info
            self._step_disabling_widget = info
            self._reaction_validation_widget = info
            self._results_enabling_widget = info
            self._tabs.update({
                "Reactive Site Filter": self._site_widget,
                "Further Exploration Filter": self._further_widget,
                "Structure Filter": self._structure_widget,
                "Elementary Step Filter": self._elementary_step_widget,
                "Reaction Filter": self._reaction_widget,
                "Aggregate Enabling": self._aggregate_enabling_widget,
                "Aggregate Validation": self._aggregate_validation_widget,
                "Reaction Enabling": self._reaction_enabling_widget,
                "Results Enabling": self._results_enabling_widget,
                "Reaction Validation": self._reaction_validation_widget,
                "Reaction Disabling": self._reaction_disabling_widget,
                "Step Disabling": self._step_disabling_widget
            })
        self._aggregate_filter_only = aggregate_filter_only
        self.set_tabs(self._tabs)

    def get_states(self) -> List[SelectionState]:
        return [widget.get_state() for widget in self._tabs.values() if isinstance(widget, ClassSelectionWidget)]

    def set_states(self, states: List[SelectionState]) -> None:
        if not all(isinstance(s, SelectionState) for s in states):
            write_error_message("Could not load filters")
            return
        widgets = [widget for widget in self._tabs.values() if isinstance(widget, ClassSelectionWidget)]
        # restart
        for widget in widgets:
            widget.restart()
        # set states
        for state, widget in zip(states[:len(widgets)], widgets):
            widget.set_state(state)

    def _save(self) -> None:
        filename = get_save_file_name(self, "filters", ["json", "pkl", "pickle"])
        if filename is None:
            write_info_message("Aborted saving")
            return
        write_info_message("Writing to disk")
        states = self.get_states()
        if filename.suffix == ".json":
            with open(filename, "w") as f:
                f.write(encode(states))
        else:
            with open(filename, "wb") as f:
                pickle.dump(states, f)

    def _load(self) -> None:
        filename = get_load_file_name(self, "filters", ["json", "pkl", "pickle"])
        if filename is None:
            write_info_message("Aborted loading")
            return
        write_info_message("Reading from disk")
        if filename.suffix == ".json":
            with open(filename, "r") as f:
                states = decode(f.read())
        else:
            with open(filename, "rb") as f:
                states = pickle.load(f)
        if not isinstance(states, list) or not all(isinstance(s, SelectionState) for s in states):
            write_error_message("Could not load filters")
            return
        self.set_states(states)

    def get_reaction_filter(self) -> Optional[ReactionFilter]:
        """
        Returns
        -------
        Optional[ReactionFilter]
            The instance of the built reaction filter, None if not built.
        """
        if self._reaction_widget is None:
            return None
        return self._reaction_widget.get_instance()

    def get_elementary_step_filter(self) -> Optional[ElementaryStepFilter]:
        """
        Returns
        -------
        Optional[ElementaryStepFilter]
            The instance of the built elementary step filter, None if not built.
        """
        if self._elementary_step_widget is None:
            return None
        return self._elementary_step_widget.get_instance()

    def get_aggregate_enabling(self) -> Optional[AggregateEnabling]:
        """
        Returns
        -------
        Optional[AggregateEnabling]
            The instance of the built aggregate enabling policy, None if not built.
        """
        if self._aggregate_enabling_widget is None:
            return None
        return self._aggregate_enabling_widget.get_instance()

    def get_reaction_enabling(self) -> Optional[ReactionEnabling]:
        """
        Returns
        -------
        Optional[ReactionEnabling]
            The instance of the built reaction enabling policy, None not if not built.
        """
        if self._reaction_enabling_widget is None:
            return None
        return self._reaction_enabling_widget.get_instance()

    def get_aggregate_validation(self) -> Optional[AggregateFilter]:
        """
        Returns
        -------
        Optional[AggregateFilter]
            The instance of the built aggregate validation filter, None not if not built.
        """
        if self._aggregate_validation_widget is None:
            return None
        return self._aggregate_validation_widget.get_instance()

    def get_reaction_validation(self) -> Optional[ReactionFilter]:
        """
        Returns
        -------
        Optional[ReactionFilter]
            The instance of the built reaction validation filter, None if not built.
        """

        if self._reaction_validation_widget is None:
            return None
        return self._reaction_validation_widget.get_instance()

    def get_results_enabling(self) -> Optional[EnableCalculationResults]:
        """
        Returns
        -------
        Optional[EnableCalculationResults]
            The instance of the built results enabling policy, None not if not built.
        """
        if self._results_enabling_widget is None:
            return None
        return self._results_enabling_widget.get_instance()

    def get_reaction_disabling(self) -> Optional[ReactionDisabling]:
        """
        Returns
        -------
        Optional[ReactionDisabling]
            The reaction disabling policy, None not if not built.
        """
        if self._reaction_disabling_widget is None:
            return None
        return self._reaction_disabling_widget.get_instance()

    def get_step_disabling(self) -> Optional[StepDisabling]:
        """
        Returns
        -------
        Optional[StepDisabling]
            The step disabling policy, None not if not built.
        """
        if self._step_disabling_widget is None:
            return None
        return self._step_disabling_widget.get_instance()

    def get_aggregate_filter(self) -> Optional[AggregateFilter]:
        """
        Returns
        -------
        Optional[AggregateFilter]
            The instance of the built aggregate filter, None if currently not valid built
        """
        return self._agg_widget.get_instance()

    def get_site_filter(self) -> Optional[ReactiveSiteFilter]:
        """
        Returns
        -------
        Optional[ReactiveSiteFilter]
            The instance of the built reactive site filter, None if currently not valid built
        """
        if self._site_widget is None:
            return None
        return self._site_widget.get_instance()

    def get_further_filter(self) -> Optional[ReactiveSiteFilter]:
        """
        Returns
        -------
        Optional[ReactiveSiteFilter]
            The instance of the built further exploration filter, None if currently not valid built
        """
        if self._further_widget is None:
            return None
        return self._further_widget.get_instance()

    def get_structure_filter(self) -> Optional[StructureFilter]:
        """
        Returns
        -------
        Optional[StructureFilter]
            The instance of the built structure filter, None if currently not valid built
        """
        if self._structure_widget is None:
            return None
        return self._structure_widget.get_instance()

    def get_filters(self) -> Tuple[AggregateFilter, ReactiveSiteFilter, FurtherExplorationFilter, StructureFilter,
                                   ElementaryStepFilter, ReactionFilter, AggregateEnabling, AggregateFilter,
                                   ReactionEnabling, EnableCalculationResults, ReactionFilter, ReactionDisabling,
                                   StepDisabling]:
        """
        Returns the filters of all tabs.
        Ensures that no None instances are returned by checking the individual instances,
        writing an error message and replacing failed instances with the default super classes.

        Returns
        -------
        Tuple[AggregateFilter, ReactiveSiteFilter, FurtherExplorationFilter, StructureFilter, ElementaryStepFilter,
                ReactionFilter, AggregateEnabling, AggregateFilter, ReactionEnabling, EnableCalculationResults,
                ReactionFilter, ReactionDisabling, StepDisabling]
            The filters of all subtabs
        """
        defaults = self.default_filters()
        if self._aggregate_filter_only:
            agg_filter = self.get_aggregate_filter()
            if agg_filter is None:
                write_error_message("AggregateFilter could not be built")
            else:
                write_info_message("Only AggregateFilter is available")
            return agg_filter, *defaults[1:]  # type: ignore

        class Sentinel:
            pass

        filters = []
        for w in self._tabs.values():
            if isinstance(w, ClassSelectionWidget):
                filters.append(w.get_instance())
            elif isinstance(w, QLineEdit):
                filters.append(Sentinel)
            else:
                filters.append(None)

        for i, _filter in enumerate(filters):
            if _filter is None:
                write_error_message(f"{defaults[i].__class__.__name__} could not be built")
                filters[i] = defaults[i]
            elif _filter is Sentinel:
                filters[i] = defaults[i]
        assert None not in filters
        filter_types = [type(d) for d in defaults]
        assert len(filters) == len(filter_types)
        # todo commented out because not all base types are implemented at the moment
        # assert all(isinstance(f, t) for f, t in zip(filters, filter_types))
        return tuple(filters)  # type: ignore

    @staticmethod
    def default_filters() -> Tuple[AggregateFilter, ReactiveSiteFilter, FurtherExplorationFilter, StructureFilter,
                                   ElementaryStepFilter, ReactionFilter, AggregateEnabling, AggregateFilter,
                                   ReactionEnabling, EnableCalculationResults, ReactionFilter, ReactionDisabling,
                                   StepDisabling]:
        return AggregateFilter(), ReactiveSiteFilter(), FurtherExplorationFilter(), StructureFilter(), \
            PlaceHolderElementaryStepFilter(), PlaceHolderReactionFilter(), PlaceHolderAggregateEnabling(), \
            PlaceHolderAggregateFilter(), PlaceHolderReactionEnabling(), PlaceHolderCalculationEnabling(), \
            PlaceHolderReactionFilter(), ReactionDisabling(), StepDisabling()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MainWindow class.
"""
import os

from pathlib import Path
from typing import Any, Optional, Union, Tuple
from PySide2.QtCore import Qt, QObject, QSettings
from PySide2.QtWidgets import QMainWindow, QApplication, QWidget
from PySide2.QtGui import QCloseEvent, QIcon

import scine_utilities as su

from scine_heron.toolbar.mo_toolbar import MOToolbar
from scine_heron.molecular_viewer import MolecularViewerWidget

from scine_heron.io.text_box import yes_or_no_question
from scine_heron.toolbar.io_toolbar import IOToolbar
from scine_heron.statusbar.status_bar import StatusBar
from scine_heron.central_widget import CentralWidget
from scine_heron.haptic.haptic_client import HapticClient
from scine_heron.mediator_potential.server_process import ServerProcess


class MainWindow(QMainWindow):
    """
    A main window with a molecule viewer and an IO toolbar.

    If file_name is not None then a molecule is loaded from this file.
    """

    def __init__(
            self,
            file_name: Optional[Path] = None,
            atom_collection: Optional[Any] = None,
            bond_orders: Optional[Any] = None,
            parent: Optional[QObject] = None,
            method_family: Optional[str] = None,
            program: Optional[str] = None,
            db_credentials: Optional[Tuple[str, str, int]] = None,
            connect_with_db: bool = False
    ):
        QMainWindow.__init__(self, parent)

        # remember some basic stuff
        self.resize(1280, 1024)
        self.settings = QSettings("Scine", "Heron")
        if self.settings.contains("size"):
            self.resize(self.settings.value("size"))  # type: ignore
        if self.settings.contains("position"):
            self.move(self.settings.value("position"))  # type: ignore

        self.setWindowTitle(self.tr("SCINE Heron"))  # type: ignore[arg-type]

        # connect haptic device
        self.__haptic_client = HapticClient()
        self.__haptic_client.init_haptic_device()
        self.__server_process = ServerProcess()

        self.__central_widget = CentralWidget(self, haptic_client=self.__haptic_client, method_family=method_family,
                                              program=program)
        self.setCentralWidget(self.__central_widget)

        self.__status_bar = StatusBar(self)
        self.__central_widget.mol_view_tab.settings_status_manager.error_update.connect(
            lambda msg: self.__status_bar.update_error_status(msg, clear_status=True)
        )
        self.__central_widget.mol_view_tab.settings_status_manager.info_update.connect(
            self.__status_bar.update_status
        )
        self.setStatusBar(self.__status_bar)

        self.toolbar = IOToolbar(self)
        self.addToolBar(Qt.TopToolBarArea, self.toolbar)
        # add function to DB connection symbol
        self.toolbar.connect_db_signal.connect(
            self.__central_widget.toggle_db_dependent_tabs
        )

        self.mo_toolbar = MOToolbar(self)
        MolecularViewerWidget.settings_widget_layout.setMenuBar(self.mo_toolbar)  # type: ignore

        # add function to load button
        self.mo_toolbar.load_file_signal.connect(self.display_molecule_from_file)
        # reset energy when molecule is loaded
        self.mo_toolbar.load_file_signal.connect(
            self.__central_widget.mol_view_tab.energy_profile_widget.reset
        )
        # add function to save button
        self.mo_toolbar.save_file_signal.connect(
            self.__central_widget.mol_view_tab.save_molecule
        )
        # add function to save button
        self.mo_toolbar.save_trajectory_signal.connect(
            self.__central_widget.mol_view_tab.save_trajectory
        )
        # add function to go back
        self.mo_toolbar.back_signal.connect(
            self.__central_widget.mol_view_tab.revert_frames
        )
        # add function to history button
        self.mo_toolbar.display_history.connect(
            self.__central_widget.mol_view_tab.show_history
        )
        # change update button symbol when signal is emitted
        self.mo_toolbar.toggle_updates_signal.connect(
            self.__continuous_updates_sanity_check
        )
        self.__central_widget.mol_view_tab.file_loaded.changed_signal.connect(
            self.mo_toolbar.show_file_loaded
        )

        # end continuous functions when quitting
        QApplication.instance().aboutToQuit.connect(
            self.__central_widget.mol_view_tab.deactivate_automatic_updates
        )
        QApplication.instance().aboutToQuit.connect(
            self.__central_widget.mol_view_tab.stop_trajectory_writing
        )
        QApplication.instance().aboutToQuit.connect(
            self.__save_settings
        )
        QApplication.instance().aboutToQuit.connect(
            lambda: self.__central_widget.chemoton_tab.closeEvent(QCloseEvent())
        )
        QApplication.instance().aboutToQuit.connect(
            lambda: self.__central_widget.steering_tab.closeEvent(QCloseEvent())
        )
        QApplication.instance().aboutToQuit.connect(
            self.__exit_handling
        )

        if file_name is not None:
            self.display_molecule_from_file(file_name)
        if atom_collection is not None:
            self.display_molecule_from_atom_collection(atom_collection, bond_orders)
        if db_credentials is not None:
            self.set_database_credentials(*db_credentials)
            if connect_with_db:
                self.toolbar.db_connection(silently=True)

    def __continuous_updates_sanity_check(self):
        value = self.__central_widget.mol_view_tab.basic_settings_widget.is_enabled()
        if value and self.__central_widget.mol_view_tab.basic_settings_widget.get_mediator_active():
            calc = self.__central_widget.mol_view_tab.create_calculator_widget.get_calculator()
            if su.Property.AtomicHessians not in calc.get_possible_properties():
                warning = "The current calculator does not support atomic Hessian, " \
                    + "hence the mediator potential requires a full Hessian, which might be slower, " \
                    + "or depending on the calculator even give wrong results. " \
                    + "Do you want to activate the updates still"
                activate = yes_or_no_question(self, warning)
                if not activate:
                    return
        # execute all the things connected with the automatic updates
        self.__update_server_process()
        self.__central_widget.mol_view_tab.toggle_automatic_updates()
        self.mo_toolbar.show_updates_enabled(value)
        self.__central_widget.mol_view_tab.configure_main_widget_updates(value)
        self.__central_widget.mol_view_tab.basic_settings_widget.set_enabled(not value)
        self.__central_widget.mol_view_tab.edit_molecule_widget.set_enabled(not value)
        self.__central_widget.mol_view_tab.create_calculator_widget.toggle_editable()

    def __save_settings(self) -> None:
        self.settings.setValue("size", self.size())
        self.settings.setValue("position", self.pos())

    def __update_server_process(self) -> None:
        """
        If the server is running, stop it, otherwise start it.
        """
        if self.__server_process.is_server_running:
            self.__server_process.stop()
            self.__server_process.is_server_running = False
        else:
            self.__server_process.start()
            self.__server_process.is_server_running = True

    def display_molecule_from_file(self, file: Path) -> None:
        """
        Imports a xyz file and displays the contained molecule.
        """
        self.__central_widget.mol_view_tab.create_molecule_widget(
            file=file,
        )

    def display_molecule_from_atom_collection(
            self, atom_collection: Any, bond_orders: Optional[Any] = None
    ) -> None:
        """
        Imports an AtomCollection and displays it.
        """
        self.__central_widget.mol_view_tab.create_molecule_widget(
            atom_collection=atom_collection,
            bond_orders=bond_orders,
        )

    def set_database_credentials(self, name: str, ip: str, port: int) -> None:
        self.toolbar.generate_db_manager(name, ip, port)

    def reject(self) -> None:
        self.__exit_handling()
        super().reject()

    def closeEvent(self, event: QCloseEvent) -> None:
        self.__exit_handling()
        super().closeEvent(event)
        QApplication.instance().closeAllWindows()

    def __exit_handling(self):
        # Stop automatic updates of the molecule
        self.__server_process.terminate()
        # Stop to save trajectory frames
        self.__central_widget.mol_view_tab.stop_trajectory_writing()
        # Disconnect haptic device
        self.__haptic_client.exit_haptic_device()

    def setIcon(self):
        dirname = os.path.dirname(__file__)
        filename = os.path.join(dirname, 'resources', 'heron_logo.png')
        appIcon = QIcon(filename)
        self.setWindowIcon(appIcon)

    def get_tab(self, tab_identifier: str) -> Union[QWidget, None]:
        if tab_identifier == 'database_viewer':
            return self.__central_widget.db_viewer_tab
        elif tab_identifier == 'readuct':
            return self.__central_widget.readuct_tab
        elif tab_identifier == 'network_viewer':
            return self.__central_widget.db_network_tab
        elif tab_identifier == 'molecule_viewer':
            return self.__central_widget.mol_view_tab
        elif tab_identifier == 'autocas':
            return self.__central_widget.autocas_tab
        elif tab_identifier == 'chemoton':
            return self.__central_widget.chemoton_tab
        elif tab_identifier == 'steering':
            return self.__central_widget.steering_tab
        return None

    def get_reaction_template_storage(self):
        return self.toolbar.template_storage

    def get_status_bar(self) -> StatusBar:
        return self.__status_bar



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from .orbital_groups import OrbitalGroup, OrbitalGroupMap


class OrbitalGroupFileReader:
    @staticmethod
    def read_orbital_group_file(file_name: str):
        orbital_index_shift = 1
        lines = open(file_name, "r").readlines()
        lines.pop(0)
        n_systems = OrbitalGroupFileReader.__get_n_systems(lines[0])
        orbital_group = OrbitalGroup(n_systems)
        map = OrbitalGroupMap([])
        for line in lines:
            n_indices = len(line.split())
            if not n_indices and not orbital_group.empty():
                orbital_group = OrbitalGroup(n_systems)
                map.add_orbital_group(orbital_group)
                continue
            orbital_group.add_orbitals([int(i) + orbital_index_shift for i in line.split()])
        map.add_orbital_group(orbital_group)
        return map

    @staticmethod
    def __get_n_systems(line):
        return len(line.split())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional

from PySide2.QtCore import QObject

from scine_chemoton.utilities.db_object_wrappers.thermodynamic_properties import ReferenceState

from scine_heron.containers.combo_box import BaseBox
from scine_heron.containers.layouts import HorizontalLayout
from scine_heron.containers.without_wheel_event import NoWheelDoubleSpinBox


class ReferenceStateBuilder(BaseBox):
    """
    Construction of a ReferenceState object, defining temperature and pressure.
    """

    def __init__(self, options: ReferenceState, parent: Optional[QObject] = None):
        super().__init__(parent=parent)
        maximum = 1e+9
        self._temperature_edit = NoWheelDoubleSpinBox(parent=parent)
        self._temperature_edit.setMinimum(0.0)
        self._temperature_edit.setMaximum(maximum)
        self._temperature_edit.setSingleStep(0.01)
        self._temperature_edit.setDecimals(2)
        self._temperature_edit.setValue(options.temperature)

        self._pressure_edit = NoWheelDoubleSpinBox(parent=parent)
        self._pressure_edit.setMinimum(0.0)
        self._pressure_edit.setMaximum(maximum)
        self._pressure_edit.setSingleStep(0.01)
        self._pressure_edit.setDecimals(2)
        self._pressure_edit.setValue(options.pressure)

        layout = HorizontalLayout()
        layout.addWidget(self._temperature_edit)
        layout.addWidget(self._pressure_edit)
        self.setLayout(layout)

    def get_reference_state(self) -> ReferenceState:
        return ReferenceState(self._temperature_edit.value(),
                              self._pressure_edit.value())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the Settings class.
"""
from typing import Optional, List
from enum import Enum


class MoleculeStyle(Enum):
    """
    Provide molecule style settings.
    """

    BallAndStick = "ball and stick"
    VDWSpheres = "VDW spheres"
    LiquoriceStick = "liquorice stick"
    Fast = "fast"
    PartialCharges = "partial charges"
    ForceArrow = "force arrow"


class LabelsStyle(Enum):
    """
    Provide labels style settings.
    """

    Empty = "empty"
    AtomicNumber = "atomic number"
    Symbol = "symbol"
    IndexNumber = "index number"


class MolViewSettings:
    """
    Provide settings for calculators and related quantities.

    calculator_settings defines the settings for running the backend.
    molecule_style defines the appearance of the molecule.
    """

    def __init__(self) -> None:
        self.molecule_style = MoleculeStyle.BallAndStick
        self.labels_style = LabelsStyle.Empty
        self.selected_molecular_orbital: Optional[int] = None
        self.number_of_molecular_orbital: Optional[int] = None
        self.molecular_orbital_value: float = 0.05
        self.mouse_picked_atom_ids: Optional[List[int]] = []
        self.haptic_picked_atom_id: Optional[int] = None
        self.force_scaling: float = 1.0
        self.error_message: str = ""
        self.info_message: str = ""
        self.haptic_force_scaling: float = 1.0
        self.gradients_scaling: float = 0.1
        self.bond_display: str = "distance"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the SettingsStatusManager class.
"""

from typing import Optional, Dict, Any, List, Tuple, TYPE_CHECKING
from PySide2.QtCore import QObject
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal

from scine_utilities import Settings

from scine_heron.settings.settings import (
    MolViewSettings,
    MoleculeStyle,
    LabelsStyle
)


class SettingsStatusManager(QObject):
    """
    A class that connects changes in the settings to signals that can be connected.
    """

    error_update = Signal(str)
    info_update = Signal(str)

    opaque_signal = Signal(list, bool)
    molecule_style_changed = Signal()
    labels_style_changed = Signal()
    selected_mo_changed = Signal(int)
    number_of_mos_changed = Signal()
    mouse_picked_atom_ids_changed = Signal()
    haptic_picked_atom_ids_changed = Signal()
    haptic_force_scaling_changed = Signal(float)
    gradients_scaling_changed = Signal(float)
    force_scaling_changed = Signal(float)

    molecular_charge_changed = Signal(int)
    spin_multiplicity_changed = Signal(int)
    spin_mode_changed = Signal(str)
    bond_display_changed = Signal(str)

    hamiltonian_changed = Signal()

    def __init__(self, parent: Optional[QObject] = None, settings_change_signal: Optional[Signal] = None):
        super().__init__(parent)
        self.__mol_view_settings = MolViewSettings()
        self.__settings_change_signal = settings_change_signal
        self.__calculator_method_family = ""
        self.__calculator_program = ""
        self.__calculator_settings: Optional[Settings] = None
        if self.__settings_change_signal is not None:
            self.__settings_change_signal.connect(self.__update_calculator_properties)  # pylint: disable=no-member
        self.__mediator_potential_active = True
        self.haptic_force_scaling_changed.connect(self.hamiltonian_changed)  # pylint: disable=no-member
        self.gradients_scaling_changed.connect(self.hamiltonian_changed)  # pylint: disable=no-member

    def get_calculator_args(self) -> Tuple[str, str]:
        return self.__calculator_method_family, self.__calculator_program

    def get_calculator_settings(self) -> Dict[str, Any]:
        if self.__calculator_settings is None:
            return {}
        return self.__calculator_settings.as_dict()

    def __update_calculator_properties(self, method_family: str, program: str, settings: Settings):
        self.__calculator_method_family = method_family
        self.__calculator_program = program
        self.__calculator_settings = settings
        if '/' in method_family and "qm_atoms" in settings:
            self.opaque_signal.emit(settings["qm_atoms"], True)
        self.hamiltonian_changed.emit()

    @property
    def error_message(self) -> str:
        return self.__mol_view_settings.error_message

    @error_message.setter
    def error_message(self, value: str) -> None:
        last_error = self.__mol_view_settings.error_message
        self.__mol_view_settings.error_message = value
        if value and value != last_error:
            self.error_update.emit(value)
        if not value and last_error:
            self.error_update.emit("Error was resolved")

    @property
    def info_message(self) -> str:
        return self.__mol_view_settings.info_message

    @info_message.setter
    def info_message(self, value: str) -> None:
        if value == self.__mol_view_settings.info_message or value == "":
            return

        self.__mol_view_settings.info_message = value
        self.info_update.emit(value)

    @property
    def molecule_style(self) -> MoleculeStyle:
        return self.__mol_view_settings.molecule_style

    @molecule_style.setter
    def molecule_style(self, value: MoleculeStyle) -> None:
        if value == self.__mol_view_settings.molecule_style:
            return
        self.__mol_view_settings.molecule_style = value
        self.molecule_style_changed.emit()

    @property
    def labels_style(self) -> LabelsStyle:
        return self.__mol_view_settings.labels_style

    @labels_style.setter
    def labels_style(self, value: LabelsStyle) -> None:
        if value == self.__mol_view_settings.labels_style:
            return
        self.__mol_view_settings.labels_style = value
        self.labels_style_changed.emit()

    @property
    def selected_molecular_orbital(self) -> Optional[int]:
        return self.__mol_view_settings.selected_molecular_orbital

    @selected_molecular_orbital.setter
    def selected_molecular_orbital(self, value: Optional[int]) -> None:
        self.__mol_view_settings.selected_molecular_orbital = value
        if not value:
            self.selected_mo_changed.emit(0)
        else:
            self.selected_mo_changed.emit(value)

    @property
    def number_of_molecular_orbital(self) -> Optional[int]:
        return self.__mol_view_settings.number_of_molecular_orbital

    @number_of_molecular_orbital.setter
    def number_of_molecular_orbital(self, value: Optional[int]) -> None:
        if value == self.__mol_view_settings.number_of_molecular_orbital:
            return
        self.__mol_view_settings.number_of_molecular_orbital = value
        self.number_of_mos_changed.emit()

    @property
    def mouse_picked_atom_ids(self) -> Optional[List[int]]:
        return self.__mol_view_settings.mouse_picked_atom_ids

    @mouse_picked_atom_ids.setter
    def mouse_picked_atom_ids(self, values: Optional[List[int]]) -> None:
        if values == self.__mol_view_settings.mouse_picked_atom_ids:
            return
        self.__mol_view_settings.mouse_picked_atom_ids = values
        self.mouse_picked_atom_ids_changed.emit()

    @property
    def haptic_picked_atom_id(self) -> Optional[int]:
        return self.__mol_view_settings.haptic_picked_atom_id

    @haptic_picked_atom_id.setter
    def haptic_picked_atom_id(self, value: Optional[int]) -> None:
        if value == self.__mol_view_settings.haptic_picked_atom_id:
            return
        self.__mol_view_settings.haptic_picked_atom_id = value
        self.haptic_picked_atom_ids_changed.emit()

    @property
    def molecular_orbital_value(self) -> float:
        return self.__mol_view_settings.molecular_orbital_value

    @molecular_orbital_value.setter
    def molecular_orbital_value(self, value: float) -> None:
        if value == self.__mol_view_settings.molecular_orbital_value:
            return
        self.__mol_view_settings.molecular_orbital_value = value
        self.selected_mo_changed.emit(value)

    def get_mediator_potential_setting(self) -> bool:
        return self.__mediator_potential_active

    def set_mediator_potential_setting(self, value: bool) -> None:
        self.__mediator_potential_active = value

    @property
    def haptic_force_scaling(self) -> float:
        return self.__mol_view_settings.haptic_force_scaling

    @haptic_force_scaling.setter
    def haptic_force_scaling(self, value: float) -> None:
        if value == self.__mol_view_settings.haptic_force_scaling:
            return
        self.__mol_view_settings.haptic_force_scaling = value
        self.haptic_force_scaling_changed.emit(value)

    @property
    def gradients_scaling(self) -> float:
        return self.__mol_view_settings.gradients_scaling

    @gradients_scaling.setter
    def gradients_scaling(self, value: float) -> None:
        if value == self.__mol_view_settings.gradients_scaling:
            return
        self.__mol_view_settings.gradients_scaling = value
        self.gradients_scaling_changed.emit(value)

    @property
    def bond_display(self) -> str:
        return self.__mol_view_settings.bond_display

    @bond_display.setter
    def bond_display(self, value: str) -> None:
        if value == self.__mol_view_settings.bond_display:
            return
        self.__mol_view_settings.bond_display = value
        self.bond_display_changed.emit(value)
